#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric_cf.actor.core.util.rpc_exception import RPCException
from fabric_cf.actor.core.apis.abc_actor_event import ABCActorEvent

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC


class FailedRPCEvent(ABCActorEvent):
    """
    Represents Failed RPC Event
    """
    def __init__(self, *, actor: ABCActorMixin, failed: FailedRPC):
        self.actor = actor
        self.failed = failed

    def process(self):
        """
        Process Failed RPC Event
        """
        self.actor.get_logger().debug(f"Processing failed RPC ({self.failed.get_request_type()})")
        if self.failed.get_request_type() == RPCRequestType.Query or \
                self.failed.get_request_type() == RPCRequestType.QueryResult or \
                self.failed.get_request_type() == RPCRequestType.ClaimDelegation or \
                self.failed.get_request_type() == RPCRequestType.ReclaimDelegation or \
                self.failed.get_request_type() == RPCRequestType.Ticket or \
                self.failed.get_request_type() == RPCRequestType.ExtendTicket or \
                self.failed.get_request_type() == RPCRequestType.Relinquish or \
                self.failed.get_request_type() == RPCRequestType.Redeem or \
                self.failed.get_request_type() == RPCRequestType.ExtendLease or \
                self.failed.get_request_type() == RPCRequestType.ModifyLease:
            self.actor.get_logger().error(f"Failed RPC for {self.failed.get_request_type()}: "
                                          f"in FailedRPCEvent.process()")
        elif self.failed.get_request_type() == RPCRequestType.Close or \
                self.failed.get_request_type() == RPCRequestType.UpdateTicket or \
                self.failed.get_request_type() == RPCRequestType.UpdateLease:
            rid = self.failed.get_reservation_id()
            if rid is None:
                self.actor.get_logger().error("Could not process failed RPC: reservation id is null")
            else:
                self.actor.handle_failed_rpc(rid=rid, rpc=self.failed)
        else:
            raise RPCException(message="Unsupported RPC request type: {}".format(
                self.failed.get_request().get_request_type()))



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server import util


class PoaPostDataVcpuCpuMap(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, vcpu: str=None, cpu: str=None):  # noqa: E501
        """PoaPostDataVcpuCpuMap - a model defined in Swagger

        :param vcpu: The vcpu of this PoaPostDataVcpuCpuMap.  # noqa: E501
        :type vcpu: str
        :param cpu: The cpu of this PoaPostDataVcpuCpuMap.  # noqa: E501
        :type cpu: str
        """
        self.swagger_types = {
            'vcpu': str,
            'cpu': str
        }

        self.attribute_map = {
            'vcpu': 'vcpu',
            'cpu': 'cpu'
        }
        self._vcpu = vcpu
        self._cpu = cpu

    @classmethod
    def from_dict(cls, dikt) -> 'PoaPostDataVcpuCpuMap':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The poa_post_data_vcpu_cpu_map of this PoaPostDataVcpuCpuMap.  # noqa: E501
        :rtype: PoaPostDataVcpuCpuMap
        """
        return util.deserialize_model(dikt, cls)

    @property
    def vcpu(self) -> str:
        """Gets the vcpu of this PoaPostDataVcpuCpuMap.


        :return: The vcpu of this PoaPostDataVcpuCpuMap.
        :rtype: str
        """
        return self._vcpu

    @vcpu.setter
    def vcpu(self, vcpu: str):
        """Sets the vcpu of this PoaPostDataVcpuCpuMap.


        :param vcpu: The vcpu of this PoaPostDataVcpuCpuMap.
        :type vcpu: str
        """

        self._vcpu = vcpu

    @property
    def cpu(self) -> str:
        """Gets the cpu of this PoaPostDataVcpuCpuMap.


        :return: The cpu of this PoaPostDataVcpuCpuMap.
        :rtype: str
        """
        return self._cpu

    @cpu.setter
    def cpu(self, cpu: str):
        """Sets the cpu of this PoaPostDataVcpuCpuMap.


        :param cpu: The cpu of this PoaPostDataVcpuCpuMap.
        :type cpu: str
        """

        self._cpu = cpu



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.manage.proxy_protocol_descriptor import ProxyProtocolDescriptor
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.apis.abc_management_object import ABCManagementObject

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.result_avro import ResultAvro


class ManagementObject(ABCManagementObject):
    """
    Base class for all manager objects. A manager object is part of the management layer.
    It provides a set of management operations for a given component, for example, actor or slice.
    Each manager object is registered with the management layer under a unique
    identifier. The creator of the object is responsible for assigning the
    correct identifier. The default constructor of ManagerObject
    generates a unique identifier for each new instance. This identifier can be
    modified and replaced with the desired identifier. Once a manager object has
    been registered with the management layer, its identifier cannot change.
 
    The type_id field of each ManagerObject can be used to
    assign the same identifier to all instances of a given manager object class.
    The type identifier field can then be used to construct an appropriate proxy
    to the ManagerObject.
 
    Each ManagerObject can be accessed using a number of protocols.
    By default, each object can be accessed using local communication.
    In addition to local communication, the object may support remote
    communication protocols such as KAFKA. Each ManagerObject maintains an array of protocol descriptors for
    each supported protocol.
 
    Each ManagerObject is responsible for its own persistence. The
    save() method is going to be invoked only once (when the
    object is registered). The reset method is going to be invoked
    every time the system has to reinstantiate the ManagerObject.
 
    A ManagerObject can be associated with a given actor. A
    ManagerObject is associated with an actor if
    ManagementObject#getActorName() returns a valid actor name. All manager objects not
    associated with an actor are considered "container-level", even though they
    may interact with one or more actors.
    """

    def __init__(self):
        self.type_id = None
        self.proxies = None
        self.id = ID()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.initialized = False
        self.serial = None

    def register_protocols(self):
        return

    def initialize(self):
        """
        Performs initialization of the manager object
        @throws Exception in case of error
        """
        if not self.initialized:
            self.register_protocols()
            if self.serial is not None:
                self.recover()

            self.initialized = True

    def save(self) -> dict:
        properties = {
                        Constants.PROPERTY_CLASS_NAME: ManagementObject.__name__,
                        Constants.PROPERTY_MODULE_NAME: ManagementObject.__module__,
                        Constants.PROPERTY_ID: str(self.id)
                     }

        if self.type_id is not None:
            properties[Constants.PROPERTY_TYPE_ID] = str(self.type_id)

        self.save_protocols(properties=properties)

        if self.get_actor_name() is not None:
            properties[Constants.PROPERTY_ACTOR_NAME] = self.get_actor_name()

        return properties

    def save_protocols(self, *, properties: dict) -> dict:
        if self.proxies is not None:
            properties[Constants.PROPERTY_PROXIES_LENGTH] = len(self.proxies)
            i = 0
            for p in self.proxies:
                properties[
                    Constants.PROPERTY_PROXIES_PREFIX + str(i) + Constants.PROPERTY_PROXIES_PROTOCOL] = p.get_protocol()
                properties[
                    Constants.PROPERTY_PROXIES_PREFIX + str(i) + Constants.PROPERTY_PROXIES_CLASS] = p.get_proxy_class()
                properties[
                    Constants.PROPERTY_PROXIES_PREFIX + str(i) + Constants.PROPERTY_PROXIES_MODULE] = p.get_proxy_module()
                i += 1
        return properties

    def load_protocols(self, *, properties: dict):
        if Constants.PROPERTY_PROXIES_LENGTH in properties:
            count = int(properties[Constants.PROPERTY_PROXIES_LENGTH])
            self.proxies = []
            for i in range(count):
                proxy = ProxyProtocolDescriptor()
                proxy.set_protocol(protocol=properties[
                    Constants.PROPERTY_PROXIES_PREFIX + str(i) + Constants.PROPERTY_PROXIES_PROTOCOL])
                proxy.set_proxy_class(proxy_class=properties[
                    Constants.PROPERTY_PROXIES_PREFIX + str(i) + Constants.PROPERTY_PROXIES_CLASS])
                proxy.set_proxy_module(proxy_module=properties[
                    Constants.PROPERTY_PROXIES_PREFIX + str(i) + Constants.PROPERTY_PROXIES_MODULE])

    def reset(self, *, properties: dict):
        self.id = ID(uid=properties[Constants.PROPERTY_ID])

        if Constants.PROPERTY_TYPE_ID in properties:
            self.type_id = ID(uid=properties[Constants.PROPERTY_TYPE_ID])

        self.load_protocols(properties=properties)

        self.serial = properties

    def recover(self):
        """
        Performs recovery actions for this manager object.
        @throws Exception in case of error
        """

    def get_id(self) -> ID:
        return self.id

    def get_actor_name(self) -> str:
        return None

    def get_type_id(self) -> ID:
        return self.type_id

    def get_proxies(self) -> list:
        return self.proxies

    @staticmethod
    def set_exception_details(*, result: ResultAvro, e: Exception):
        result.message += str(e)
        result.set_details(traceback.format_exc())
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest
import time
from fabric_cf.actor.core.apis.abc_container_database import ABCContainerDatabase
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.test.base_test_case import BaseTestCase


class ContainerDatabaseTest(BaseTestCase, unittest.TestCase):
    from fabric_cf.actor.core.container.globals import Globals
    Globals.config_file = "./config/config.test.yaml"
    Constants.SUPERBLOCK_LOCATION = './state_recovery.lock'

    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    container = GlobalsSingleton.get().get_container()

    def get_clean_database(self) -> ABCContainerDatabase:
        self.db = self.container.get_database()
        return self.db

    def test_a_create(self):
        db = self.get_clean_database()
        self.assertEqual(1, len(db.get_actors()), "No actors")
        self.assertIsNotNone(db.get_container_properties())
        self.assertIsNotNone(db.get_time())

    def test_b_add_remove_actor(self):
        db = self.get_clean_database()

        self.get_actor()

        result = db.get_actor(actor_name=self.actor_name)
        print(result)
        self.assertIsNotNone(result)

        result = db.get_actors()

        self.assertIsNotNone(result)
        self.assertEqual(2, len(result), "Two actor")

        db.remove_actor(actor_name=self.actor_name)

        self.assertIsNone(db.get_actor(actor_name=self.actor_name))

        self.assertEqual(1, len(db.get_actors()), "One actor")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING

from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_database import ABCDatabase
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice


class ABCBasePlugin(ABC):
    """
    IActorPlugin defines the interface for linking/injecting functionality to the leasing code.
    This interface can be used to link other systems to the core, for example to add support for leasing
    to a cluster management system.

    These methods are called as various events occur. All implementations of this class must have a
    constructor that takes no arguments, and set methods for their attributes.
    """
    @abstractmethod
    def configure(self, *, properties):
        """
        Processes a list of configuration properties. This method is called by the configuration engine.

        Args:
            properties: properties

        Raises:
            Exception in case of error
        """

    @abstractmethod
    def actor_added(self, *, config: ActorConfig):
        """
        Performs initialization steps that require that the actor has been added

        Raises:
            Exception in case of error
        """

    @abstractmethod
    def recovery_starting(self):
        """
        Informs the plugin that recovery is about to start.
        """

    @abstractmethod
    def initialize(self):
        """
        Initializes the actor. Called early in the initialization process.

        Raises:
            Exception in case of error
        """

    @abstractmethod
    def revisit(self, *, slice_obj: ABCSlice = None, reservation: ABCReservationMixin = None, delegation: ABCDelegation = None):
        """
        Rebuilds plugin state associated with a restored slice/reservation. Called once for each restored
        slice/reservation.

        Args:
            slice_obj: restored slice
            reservation: restored reservation
            delegation: restored delegation
        Raises:
            Exception if rebuilding state fails
        """

    @abstractmethod
    def recovery_ended(self):
        """
        Informs the plugin that recovery has completed.
        """

    @abstractmethod
    def restart_configuration_actions(self, *, reservation: ABCReservationMixin):
        """
        Restarts any pending configuration actions for the specified reservation

        Args:
            reservation: reservation
        Raises:
            Exception if restarting actions fails
        """

    @abstractmethod
    def create_slice(self, *, slice_id: ID, name: str, project_id: str, project_name: str):
        """
        Creates a new slice.

        Args:
            slice_id: guid for the slice
            name: slice name
            project_id: project id
            project_name: project_name
        Returns:
            slice object
        Raises:
            Exception if creation fails
        """

    @abstractmethod
    def release_slice(self, *, slice_obj: ABCSlice):
        """
        Releases any resources held by the slice.

        Args:
            slice_obj: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def validate_incoming(self, *, reservation: ABCReservationMixin, auth: AuthToken):
        """
        Validates an incoming reservation request

        Args:
            reservation: reservation
            auth: auth token of the caller
        Returns:
            True if the validation succeeds
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def set_actor(self, *, actor: ABCActorMixin):
        """
        Sets the actor. Note: the actor has to be fully initialized.

        Args:
            actor: actor
        """

    @abstractmethod
    def get_actor(self):
        """
        Returns the actor associated with the plugin

        Returns:
            actor associated with the plugin
        """

    @abstractmethod
    def get_logger(self):
        """
        Returns the logger.

        @returns logger instance
        """

    @abstractmethod
    def get_database(self) -> ABCDatabase:
        """
        Obtains the actor's database instance.

        @return database instance
        """

    @abstractmethod
    def set_database(self, *, db: ABCDatabase):
        """
        Sets the actor's database instance.

        @param db
                   database instance
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from datetime import datetime, timezone

from fim.slivers.capacities_labels import ReservationInfo

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin, ReservationCategory
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.event_logger import EventLogger, EventLoggerSingleton
from fabric_cf.actor.core.common.exceptions import ReservationException
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates, JoinState
from fabric_cf.actor.core.proxies.kafka.translate import Translate
from fabric_cf.actor.core.util.reservation_state import ReservationState

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.kernel.request_types import RequestTypes
    from fabric_cf.actor.core.kernel.resource_set import ResourceSet
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.util.resource_type import ResourceType
    from fabric_cf.actor.core.kernel.poa import Poa


class Reservation(ABCReservationMixin):
    """
    These are the only methods synchronized on the Reservation object itself. The
    purpose is to allow an external thread to await a state transition in a
    Reservation without holding the orchestrator lock. State changes are made only
    while holding the orchestrator lock, so a orchestrator may examine the state without
    acquiring the reservation lock.

    Reservation objects passed into a slices actor to initiate or request new
    reservations are taken over by the kernel. Validate the passed-in state, mark
    some context specific to the operation, and clean out the rest of it in
    preparation to link it into orchestrator structures. No locks are held, and these
    routines have no side effects other than to the (new) reservation.

    Reservation is the base for all reservation objects. It
    implements a part of the IReservation interface and defines the
    core functions expected by the kernel from all reservation classes. This is
    an abstract class and is intended as a building block of higher-level
    reservation classes.
    """

    def __init__(self, *, rid: ID = None, resources: ResourceSet = None, term: Term = None,
                 slice_object: ABCSlice = None):
        # The unique reservation identifier.
        self.rid = rid
        # Reservation category. Subclasses should supply the correct value.
        self.category = ReservationCategory.All
        # Reservation state.
        self.state = ReservationStates.Nascent
        # Reservation pending state.
        self.pending_state = ReservationPendingStates.None_
        # Has this reservation ever been extended?
        self.extended = False
        # The current resources associated with this reservation.
        self.resources = resources
        # Resources representing the last request issued/received for this reservation
        self.requested_resources = None
        # Resources approved by the policy for this reservation. This resource set
        # can be different from what was initially requested (requested_resources)
        # Eventually, resources will be merged with approved_resources.
        self.approved_resources = None
        # The current term of the reservation.
        self.term = term
        # The previous term of the reservation.
        self.previous_term = None
        # The term of the last request issued/received for this reservation.
        self.requested_term = None
        # The term the policy approved for this reservation. This term can be
        # different from what was initially requested (requested_term). Eventually,
        # term will be set to equal approved_term.
        self.approved_term = None
        # True if this is a renewable reservation. By default, reservations are not renewable
        self.renewable = None
        # Last error message.
        self.error_message = None
        # Cached pointer to the actor that operates on this reservation.
        self.actor = None
        # Logger
        self.logger = None
        # Slice this reservation belongs to.
        self.slice = slice_object
        if slice_object is not None:
            # Cached slice name. Necessary so that we can obtain the slice for
            # reservations that have not been fully recovered.
            self.slice_name = slice_object.get_name()
            # Cached slice id. Necessary so that we can obtain the slice for
            # reservations that have not been fully recovered.
            self.slice_id = slice_object.get_slice_id()
        else:
            self.slice_name = None
            self.slice_id = None
        # Indicates if the policy plugin has made a decision about this reservation
        self.approved = False
        # The resources assigned to the reservation before the last update.
        self.previous_resources = None
        # Is an allocation process in progress?
        self.bid_pending = False
        # Dirty flag. Indicates that the state of the reservation object has
        # changed since the last time it was persisted. Currently only transition
        # updates the dirty flag
        self.dirty = False
        # True if this reservation is expired. Used during recovery.
        self.expired = False
        # Recovery flag.
        self.pending_recover = False
        # True if the last state transition is not committed to external storage.
        # false otherwise.
        self.state_transition = False
        # Scratch element to trigger post-actions on a probe.
        self.service_pending = ReservationPendingStates.None_
        self.last_transition_time = None
        self.last_pending_state = ReservationPendingStates.None_
        self.thread_lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['actor']
        del state['logger']
        del state['slice']
        del state['approved']
        del state['previous_resources']
        del state['bid_pending']
        del state['dirty']
        del state['expired']
        del state['pending_recover']
        del state['state_transition']
        del state['service_pending']
        del state['thread_lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.actor = None
        self.logger = None
        self.slice = None
        self.approved = False
        self.previous_resources = None
        self.bid_pending = False
        self.dirty = False
        self.expired = False
        self.pending_recover = False
        self.state_transition = False
        self.service_pending = ReservationPendingStates.None_
        self.thread_lock = threading.Lock()

    def restore(self, *, actor: ABCActorMixin, slice_obj: ABCSlice):
        """
        Must be invoked after creating reservation from unpickling
        """
        self.actor = actor
        self.slice = slice_obj
        if actor is not None:
            self.logger = self.actor.get_logger()
        self.approved = False
        self.previous_resources = None
        self.bid_pending = False
        self.dirty = False
        self.expired = False
        self.pending_recover = False
        self.state_transition = False
        self.service_pending = ReservationPendingStates.None_
        if actor is not None:
            if self.resources is not None:
                self.resources.restore(plugin=actor.get_plugin(), reservation=self)
            if self.requested_resources is not None:
                self.requested_resources.restore(plugin=actor.get_plugin(), reservation=self)
            if self.approved_resources is not None:
                self.approved_resources.restore(plugin=actor.get_plugin(), reservation=self)

    def can_redeem(self) ->bool:
        return True

    def can_renew(self) -> bool:
        return True

    def internal_error(self, *, err: str):
        """
        Internal error log and raise exception
        """
        self.logger.error(f"internal error for reservation: {self} : {err}")
        raise ReservationException(f"internal error: {err}")

    def error(self, *, err: str):
        """
        Error log and raise exception
        """
        if self.logger is not None:
            self.logger.error(f"error for reservation: {self} : {err}")
        else:
            print(f"error for reservation: {self} : {err}")
        raise ReservationException(f"error: {err}")

    def clear_dirty(self):
        """
        Clear dirty flag
        """
        self.dirty = False
        self.state_transition = False

    def clear_notice(self, clear_fail: bool=False):
        """
        Clears all event notices associated with the reservation.
        """

    def close(self):
        """
        Close a reservation
        """

    def extend_lease(self):
        """
        Extend lease on reservation
        """

    def modify_lease(self):
        """
        Modify lease on reservation
        """

    def extend_ticket(self, *, actor: ABCActorMixin):
        """
        Extend a ticket
        """
        self.internal_error(err="abstract extend_ticket trap")

    def fail(self, *, message: str, exception: Exception = None):
        """
        Fail a reservation
        """
        self.error_message = message
        self.bid_pending = False
        self.transition(prefix=message, state=ReservationStates.Failed, pending=ReservationPendingStates.None_)
        self.logger.error(f"{message}  e: {exception}")

    def fail_warn(self, *, message: str):
        """
        Fail with a warning
        """
        self.error_message = message
        self.transition(prefix=message, state=ReservationStates.Failed, pending=ReservationPendingStates.None_)
        message = f"reservation has failed: {message} : [{self}]"
        self.logger.warning(message)

    def get_actor(self):
        return self.actor

    def get_approved_resources(self) -> ResourceSet:
        return self.approved_resources

    def get_approved_term(self) -> Term:
        return self.approved_term

    def get_approved_type(self) -> ResourceType:
        if self.approved_resources is not None:
            return self.approved_resources.get_type()
        return None

    def get_approved_units(self) -> int:
        if self.approved_resources is not None:
            return self.approved_resources.get_units()
        return 0

    def get_category(self) -> ReservationCategory:
        return self.category

    def get_kernel_slice(self) -> ABCSlice:
        return self.slice

    def get_leased_abstract_units(self) -> int:
        return 0

    def get_leased_units(self) -> int:
        return 0

    def get_notices(self) -> str:
        """
        Returns a descriptive string if this reservation requires attention, else None

        @return notices string
        """
        msg = f"Reservation {self.rid} (Slice {self.slice}) is in state ({self.get_state_name()}," \
              f"{self.get_pending_state_name()})"

        if self.error_message is not None and self.error_message != "":
            msg += f", err={self.error_message}"
        return msg

    def get_pending_state(self) -> ReservationPendingStates:
        return self.pending_state

    def get_pending_state_name(self) -> str:
        return ReservationPendingStates(self.pending_state).name

    def get_previous_resources(self) -> ResourceSet:
        return self.previous_resources

    def get_previous_term(self) -> Term:
        return self.previous_term

    def get_requested_resources(self) -> ResourceSet:
        return self.requested_resources

    def get_requested_term(self) -> Term:
        return self.requested_term

    def get_requested_type(self) -> ResourceType:
        if self.requested_resources is not None:
            return self.requested_resources.get_type()
        return None

    def get_requested_units(self) -> int:
        if self.requested_resources is not None:
            return self.requested_resources.get_units()
        return 0

    def get_reservation_id(self) -> ID:
        return self.rid

    def get_reservation_state(self) -> ReservationState:
        return ReservationState(state=self.state, pending=self.pending_state)

    def get_resources(self) -> ResourceSet:
        return self.resources

    def get_slice(self) -> ABCSlice:
        return self.slice

    def get_slice_id(self):
        if self.slice is None:
            return None
        return self.slice.get_slice_id()

    def get_slice_name(self):
        if self.slice is None:
            return None
        return self.slice.get_name()

    def get_state(self) -> ReservationStates:
        return self.state

    def get_state_name(self) -> str:
        return self.state.name

    def get_term(self) -> Term:
        return self.term

    def get_type(self) -> ResourceType:
        if self.resources is None:
            return None
        return self.resources.get_type()

    def get_units(self, *, when: datetime = None) -> int:
        if when is None:
            if self.resources is None:
                return 0
            return self.resources.get_units()
        if not self.is_terminal() and self.term is not None and self.term.contains(date=when):
            return self.resources.get_concrete_units(when=when)
        return 0

    def handle_duplicate_request(self, *, operation: RequestTypes):
        return

    def has_uncommitted_transition(self) -> bool:
        return self.state_transition

    def is_active(self) -> bool:
        return self.state == ReservationStates.Active or self.state == ReservationStates.ActiveTicketed

    def is_active_ticketed(self) -> bool:
        return self.state == ReservationStates.ActiveTicketed

    def is_approved(self) -> bool:
        return self.approved

    def is_bid_pending(self) -> bool:
        """
        Is bid pending
        @return bid pending
        """
        return self.bid_pending

    def is_closed(self) -> bool:
        return self.state == ReservationStates.Closed

    def is_closing(self) -> bool:
        return self.state == ReservationStates.CloseWait or self.pending_state == ReservationPendingStates.Closing

    def is_dirty(self) -> bool:
        return self.dirty

    def is_expired(self, *, t: datetime = None) -> bool:
        if t is None:
            return self.expired
        return self.term.expired(date=t)

    def is_extended(self) -> bool:
        return self.extended

    def is_extending_lease(self) -> bool:
        return self.pending_state == ReservationPendingStates.ExtendingLease

    def is_extending_ticket(self) -> bool:
        return self.pending_state == ReservationPendingStates.ExtendingTicket

    def is_failed(self) -> bool:
        return self.state == ReservationStates.Failed

    def is_nascent(self) -> bool:
        return self.state == ReservationStates.Nascent

    def is_no_pending(self) -> bool:
        return self.pending_state == ReservationPendingStates.None_

    def is_pending_recover(self):
        return self.pending_recover

    def is_priming(self) -> bool:
        return self.pending_state == ReservationPendingStates.Priming

    def is_redeeming(self) -> bool:
        return self.pending_state == ReservationPendingStates.Redeeming

    def is_renewable(self) -> bool:
        """
        Is reservation renewable
        @return true if renewable; false otherwise
        """
        return self.renewable

    def is_terminal(self) -> bool:
        return self.is_closed() or self.is_closing() or self.is_failed()

    def is_ticketed(self) -> bool:
        return self.state == ReservationStates.Ticketed

    def is_ticketing(self) -> bool:
        return self.pending_state == ReservationPendingStates.Ticketing

    def nothing_pending(self):
        """
        Ensures the reservation does not have a pending operation.

        @throws Exception if the reservation has a pending operation.
        """
        if self.pending_state != ReservationPendingStates.None_:
            self.error(err=Constants.PENDING_OPERATION_ERROR)

    def prepare_probe(self):
        return

    def probe_pending(self):
        return

    def ready(self):
        """
        An incoming client request named this validated Reservation object for an
        existing reservation. Check to be sure that it has not been destroyed in
        a race since the validate.

        @throws Exception thrown if the state is closed or failed
        """
        if self.state == ReservationStates.Closed or self.state == ReservationStates.Failed:
            self.error(err="invalid Reservation")

    def reserve(self, *, policy: ABCPolicy):
        return

    def setup(self):
        """
        Setup a reservation
        """
        if self.resources is not None:
            self.resources.setup(reservation=self)

        if self.approved_resources is not None:
            self.approved_resources.setup(reservation=self)

        if self.requested_resources is not None:
            self.requested_resources.setup(reservation=self)

    def service_claim(self):
        return

    def service_reclaim(self):
        return

    def service_close(self):
        return

    def service_extend_lease(self):
        return

    def service_modify_lease(self):
        return

    def service_extend_ticket(self):
        return

    def service_probe(self):
        return

    def service_reserve(self):
        return

    def service_update_lease(self):
        return

    def service_update_ticket(self):
        return

    def set_actor(self, *, actor: ABCActorMixin):
        self.actor = actor

    def set_approved(self, *, term: Term = None, approved_resources: ResourceSet = None):
        self.approved_term = term
        self.approved_resources = approved_resources
        self.approved = True

    def set_approved_resources(self, *, approved_resources: ResourceSet):
        self.approved_resources = approved_resources

    def set_approved_term(self, *, term: Term):
        self.approved_term = term

    def set_bid_pending(self, *, value: bool):
        """
        Set Bid Pending
        @param value value
        """
        self.bid_pending = value

    def set_dirty(self):
        """
        Set dirty
        """
        self.dirty = True

    def set_expired(self, *, value: bool):
        self.expired = value

    def set_logger(self, *, logger):
        self.logger = logger

    def set_pending_recover(self, *, pending_recover: bool):
        self.pending_recover = pending_recover

    def set_service_pending(self, *, code: ReservationPendingStates):
        self.service_pending = code

    def set_slice(self, *, slice_object: ABCSlice):
        self.slice = slice_object

    def __str__(self):
        msg = "res: "
        if self.rid is not None:
            msg += f"#{self.rid} "

        if self.slice is not None:
            msg += f"slice: [{self.slice}] "

        msg += f"state:[{self.get_state_name()},{self.get_pending_state_name()}] "

        if self.resources is not None:
            msg += f"resources: [{self.resources}] "

        if self.term is not None:
            msg += f"term: [{self.term}]"

        return msg

    def transition(self, *, prefix: str, state: ReservationStates, pending: ReservationPendingStates):
        if self.state == ReservationStates.Failed and self.logger is not None:
            self.logger.debug("failed")

        if self.logger is not None:
            self.logger.debug(f"Reservation #{self.rid} {prefix} transition: {self.get_state_name()} -> {state.name}, "
                              f"{self.get_pending_state_name()} -> {pending.name}")

        change = self.state != state
        self.state = state
        self.last_pending_state = pending
        self.pending_state = pending

        self.set_dirty()
        self.state_transition = True
        self.last_transition_time = datetime.now(timezone.utc)

        if change:
            sliver = None
            if self.get_resources() is not None:
                sliver = self.get_resources().get_sliver()
            elif self.get_requested_resources() is not None:
                sliver = self.get_requested_resources().get_sliver()
            if sliver is not None:
                if sliver.reservation_info is None:
                    sliver.reservation_info = ReservationInfo()
                sliver.reservation_info.reservation_id = str(self.get_reservation_id())
                sliver.reservation_info.reservation_state = str(self.state)
                EventLoggerSingleton.get().log_sliver_event(
                    slice_object=Translate.translate_slice_to_avro(slice_obj=self.slice), sliver=sliver)

    def update_lease(self, *, incoming: ABCReservationMixin, update_data):
        self.internal_error(err="abstract update_lease trap")

    def update_ticket(self, *, incoming: ABCReservationMixin, update_data):
        self.internal_error(err="abstract update_ticket trap")

    def validate_outgoing(self):
        return

    def validate_incoming(self):
        return

    def validate(self):
        """
        Validates the reservation. For use by prepare() methods defined by
        subclasses.

        @throws Exception
        """
        assert self.state == ReservationStates.Nascent
        self.nothing_pending()

        if self.slice is None:
            self.error(err="no slice specified")

        if self.resources is None:
            self.error(err="no resource set specified")

        if self.term is None:
            self.error(err="no term specified")

        self.term.validate()

    def get_join_state(self) -> JoinState:
        """
        Get Join State
        @return join state
        """
        return JoinState.None_

    def get_graph_node_id(self) -> str:
        if self.requested_resources is not None:
            request = self.requested_resources.get_sliver()
            if request is not None:
                node_map = request.get_node_map()
                if node_map is not None:
                    return node_map[1]
        return None

    def exceeds_timeout(self, timeout: int) -> bool:
        """
        Check if Reservation has been in the state for more than the timeout interval
        @param timeout: timeout
        @return True if reservation has been in the state for more than timeout; False otherwise
        """
        if self.last_transition_time is None:
            return False

        current_time = datetime.now(timezone.utc)
        if (current_time - self.last_transition_time).seconds > timeout:
            return True

        return False

    def get_error_message(self) -> str:
        return self.error_message

    def lock(self):
        self.thread_lock.acquire()

    def unlock(self):
        if self.thread_lock.locked():
            self.thread_lock.release()

    def service_poa(self):
        pass

    def poa_info(self, *, incoming: Poa):
        pass

    def poa(self, *, poa: Poa):
        """
        POA on reservation
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, List

from fabric_mb.message_bus.messages.actor_avro import ActorAvro
from fabric_mb.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
from fabric_mb.message_bus.messages.lease_reservation_state_avro import LeaseReservationStateAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.reservation_predecessor_avro import ReservationPredecessorAvro
from fabric_mb.message_bus.messages.reservation_state_avro import ReservationStateAvro
from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
from fabric_mb.message_bus.messages.unit_avro import UnitAvro
from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation
from fabric_cf.actor.core.apis.abc_controller_reservation import ABCControllerReservation
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.actor_identity import ActorIdentity
from fabric_cf.actor.core.core.ticket import Ticket
from fabric_cf.actor.core.core.unit import Unit
from fabric_cf.actor.core.kernel.predecessor_state import PredecessorState
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.proxies.actor_location import ActorLocation
from fabric_cf.actor.core.proxies.kafka.kafka_proxy import KafkaProxy
from fabric_cf.actor.core.proxies.local.local_proxy import LocalProxy
from fabric_cf.actor.core.util.client import Client
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.core.manage.messages.client_mng import ClientMng

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.apis.abc_proxy import ABCProxy
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin


class Converter:
    @staticmethod
    def absorb_res_properties(*, rsv_mng: ReservationMng, res_obj: ABCReservationMixin):
        res_obj.get_resources().set_sliver(sliver=rsv_mng.get_sliver())
        return res_obj

    @staticmethod
    def fill_reservation_predecessor(*, pred: PredecessorState) -> ReservationPredecessorAvro:
        pred_avro = ReservationPredecessorAvro()
        pred_avro.set_reservation_id(value=str(pred.get_reservation().get_reservation_id()))
        return pred_avro

    @staticmethod
    def fill_reservation(*, reservation: ABCReservationMixin, full: bool) -> ReservationMng:
        rsv_mng = None
        if isinstance(reservation, ABCControllerReservation):
            rsv_mng = LeaseReservationAvro()
            if reservation.get_redeem_predecessors() is not None and len(reservation.get_redeem_predecessors()) > 0:
                rsv_mng.redeem_processors = []
                for p in reservation.get_redeem_predecessors():
                    pred = Converter.fill_reservation_predecessor(pred=p)
                    rsv_mng.redeem_processors.append(pred)

        elif isinstance(reservation, ABCClientReservation):
            rsv_mng = TicketReservationAvro()
        else:
            rsv_mng = ReservationMng()

        rsv_mng.set_reservation_id(str(reservation.get_reservation_id()))
        rsv_mng.set_slice_id(str(reservation.get_slice_id()))

        if reservation.get_type() is not None:
            rsv_mng.set_resource_type(str(reservation.get_type()))

        rsv_mng.set_units(reservation.get_units())
        rsv_mng.set_state(reservation.get_state().value)
        rsv_mng.set_pending_state(reservation.get_pending_state().value)

        if isinstance(reservation, ABCControllerReservation):
            rsv_mng.set_leased_units(reservation.get_leased_abstract_units())
            rsv_mng.set_join_state(reservation.get_join_state().value)
            authority = reservation.get_authority()

            if authority is not None:
                rsv_mng.set_authority(str(authority.get_guid()))

        if isinstance(reservation, ABCClientReservation):
            broker = reservation.get_broker()
            if broker is not None:
                rsv_mng.set_broker(str(broker.get_guid()))
            rsv_mng.set_renewable(reservation.is_renewable())
            rsv_mng.set_renew_time(reservation.get_renew_time())

        if reservation.get_term() is not None:
            rsv_mng.set_start(ActorClock.to_milliseconds(when=reservation.get_term().get_start_time()))
            rsv_mng.set_end(ActorClock.to_milliseconds(when=reservation.get_term().get_end_time()))
        else:
            if reservation.get_requested_term() is not None:
                rsv_mng.set_start(ActorClock.to_milliseconds(when=reservation.get_requested_term().get_start_time()))
                rsv_mng.set_end(ActorClock.to_milliseconds(when=reservation.get_requested_term().get_end_time()))

        if reservation.get_requested_term() is not None:
            rsv_mng.set_requested_end(ActorClock.to_milliseconds(when=reservation.get_requested_term().get_end_time()))

        rsv_mng.set_notices(reservation.get_notices())

        if full:
            rsv_mng = Converter.attach_res_properties(mng=rsv_mng, reservation=reservation)

        return rsv_mng

    @staticmethod
    def attach_res_properties(*, mng: ReservationMng, reservation: ABCReservationMixin):
        sliver = None
        if isinstance(reservation, ABCControllerReservation):
            if reservation.is_active():
                sliver = reservation.get_leased_resources().get_sliver()
            else:
                sliver = reservation.get_resources().get_sliver()
        else:
            rset = reservation.get_resources()
            if rset is not None:
                sliver = rset.get_sliver()

        ticket = None
        rset = reservation.get_resources()

        if rset is not None:
            cs = rset.get_resources()

            if cs is not None and isinstance(cs, Ticket):
                ticket = cs.get_properties()

        mng.set_sliver(sliver=sliver)

        if isinstance(mng, TicketReservationAvro):
            mng.set_ticket_properties(ticket)

        return mng

    @staticmethod
    def fill_reservation_state(*, res: ABCReservationMixin) -> ReservationStateAvro:
        result = None
        if isinstance(res, ABCControllerReservation):
            result = LeaseReservationStateAvro()
            result.set_reservation_id(rid=str(res.get_reservation_id()))
            result.set_state(res.get_state().value)
            result.set_pending_state(res.get_pending_state().value)
            result.set_joining(res.get_join_state().value)
        else:
            result = ReservationStateAvro()
            result.set_reservation_id(rid=str(res.get_reservation_id()))
            result.set_state(res.get_state().value)
            result.set_pending_state(res.get_pending_state().value)

        return result

    @staticmethod
    def fill_reservation_states(*, res_list: list) -> List[ReservationStateAvro]:
        result = []
        for r in res_list:
            rstate = Converter.fill_reservation_state(res=r)
            result.append(rstate)

        return result

    @staticmethod
    def fill_client(*, client_mng: ClientMng) -> Client:
        result = Client()
        result.set_name(name=client_mng.get_name())
        result.set_guid(guid=ID(uid=client_mng.get_guid()))
        return result

    @staticmethod
    def fill_client_mng(*, client: Client) -> ClientMng:
        result = ClientMng()
        result.set_name(name=client.get_name())
        result.set_guid(guid=str(client.get_guid()))
        return result

    @staticmethod
    def fill_clients(*, client_list: List[Client]) -> List[ClientMng]:
        result = []
        for c in client_list:
            mng = Converter.fill_client_mng(client=c)
            result.append(mng)

        return result

    @staticmethod
    def fill_unit_mng(*, unit: Unit) -> UnitAvro:
        result = UnitAvro()
        result.properties = unit.properties
        return result

    @staticmethod
    def fill_units(*, unit_list: List[Unit]) -> List[UnitAvro]:
        result = []
        for u in unit_list:
            mng = Converter.fill_unit_mng(unit=u)
            result.append(mng)

        return result

    @staticmethod
    def fill_proxy(*, proxy: ABCProxy) -> ProxyAvro:
        result = ProxyAvro()
        result.set_name(proxy.get_name())
        result.set_guid(str(proxy.get_guid()))

        if isinstance(proxy, LocalProxy):
            result.set_protocol(Constants.PROTOCOL_LOCAL)
        elif isinstance(proxy, KafkaProxy):
            result.set_protocol(Constants.PROTOCOL_KAFKA)
            result.set_kafka_topic(proxy.get_kafka_topic())

        return result

    @staticmethod
    def fill_proxies(*, proxies: list) -> List[ProxyAvro]:
        result = []
        for p in proxies:
            proxy = Converter.fill_proxy(proxy=p)
            result.append(proxy)

        return result

    @staticmethod
    def get_agent_proxy(*, mng: ProxyAvro):
        try:
            location = ActorLocation(location=mng.get_kafka_topic())
            identity = ActorIdentity(name=mng.get_name(), guid=ID(uid=mng.get_guid()))
            from fabric_cf.actor.core.container.container import Container
            return Container.get_proxy(protocol=mng.get_protocol(), identity=identity, location=location,
                                       proxy_type=mng.get_type())
        except Exception:
            traceback.format_exc()
            return None

    @staticmethod
    def get_resource_set(*, res_mng: ReservationMng) -> ResourceSet:
        return ResourceSet(units=res_mng.get_units(), rtype=ResourceType(resource_type=res_mng.get_resource_type()),
                           sliver=res_mng.get_sliver())

    @staticmethod
    def fill_actor(*, actor: ABCActorMixin) -> ActorAvro:
        result = ActorAvro()
        result.set_name(actor.get_name())
        result.set_description(actor.get_description())
        result.set_type(actor.get_type().value)
        result.set_online(True)
        result.set_id(str(actor.get_guid()))
        result.set_policy_guid(str(actor.get_plugin().get_guid()))
        return result

    @staticmethod
    def fill_actors(*, act_list: list) -> List[ActorAvro]:
        result = []
        for a in act_list:
            act_mng = Converter.fill_actor(actor=a)
            result.append(act_mng)

        return result

    @staticmethod
    def fill_actor_from_db(*, actor: ABCActorMixin) -> ActorAvro:
        result = ActorAvro()
        result.set_description(actor.get_description())
        result.set_name(actor.get_name())
        result.set_type(actor.get_type().value)

        from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
        aa = ActorRegistrySingleton.get().get_actor(actor_name_or_guid=actor.get_name())
        result.set_online(aa is not None)

        return result

    @staticmethod
    def fill_actors_from_db(*, act_list: List[ABCActorMixin]) -> List[ActorAvro]:
        result = []
        for a in act_list:
            act_mng = Converter.fill_actor_from_db(actor=a)
            result.append(act_mng)

        return result

    @staticmethod
    def fill_actors_from_db_status(*, act_list: List[ABCActorMixin], status: int) -> List[ActorAvro]:
        result = []
        for a in act_list:
            act_mng = Converter.fill_actor_from_db(actor=a)

            if status == 0 or (status == 1 and act_mng.get_online()) or (status == 2 and not act_mng.get_online()):
                result.append(act_mng)

        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import random
import traceback
from typing import Tuple, List, Union

from fim.graph.abc_property_graph import ABCPropertyGraph, ABCGraphImporter
from fim.graph.neo4j_property_graph import Neo4jGraphImporter, Neo4jPropertyGraph
from fim.graph.networkx_property_graph import NetworkXGraphImporter
from fim.graph.resources.abc_arm import ABCARMPropertyGraph
from fim.graph.resources.abc_cbm import ABCCBMPropertyGraph
from fim.graph.resources.neo4j_arm import Neo4jARMGraph
from fim.graph.resources.neo4j_cbm import Neo4jCBMGraph, Neo4jCBMFactory
from fim.graph.slices.abc_asm import ABCASMPropertyGraph
from fim.graph.slices.neo4j_asm import Neo4jASMFactory
from fim.graph.slices.networkx_asm import NetworkxASM, NetworkXASMFactory
from fim.slivers.attached_components import ComponentSliver
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.capacities_labels import Capacities
from fim.slivers.delegations import Delegations
from fim.slivers.interface_info import InterfaceSliver, InterfaceType
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver, ServiceType
from fim.user import ExperimentTopology, Labels, NodeType, Component, ReservationInfo
from fim.user.interface import Interface

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates


class InterfaceSliverMapping:
    def __init__(self):
        # Maps to Connection Point in the Graph
        self.peer_ifs = None
        # Maps to Network Service (Parent of Peer IFS) in the Graph
        self.peer_ns_id = None
        # Maps to Component (Parent of Network Service) in the Graph
        self.component_name = None
        # Maps to the Network Node (Parent of the Component or Parent of Connection) in the Graph
        self.node_id = None
        self.facility = False
        self.peer_site = None

    def get_peer_ifs(self) -> InterfaceSliver:
        return self.peer_ifs

    def get_peer_ns_id(self) -> str:
        return self.peer_ns_id

    def get_component_name(self) -> str:
        return self.component_name

    def get_node_id(self) -> str:
        return self.node_id

    def is_facility(self) -> bool:
        return self.facility

    def get_peer_site(self) -> str:
        return self.peer_site

    def is_peered(self) -> bool:
        return self.peer_site is not None

    def set_peer_ifs(self, peer_ifs: InterfaceSliver):
        self.peer_ifs = peer_ifs

    def set_peer_ns_id(self, peer_ns_id: str):
        self.peer_ns_id = peer_ns_id

    def set_component_name(self, component_name: str):
        self.component_name = component_name

    def set_node_id(self, node_id: str):
        self.node_id = node_id

    def set_facility(self, facility: bool):
        self.facility = facility

    def set_peer_site(self, peer_site: str):
        self.peer_site = peer_site

    def set_properties(self, **kwargs):
        """
        Lets you set multiple properties exposed via setter methods
        :param kwargs:
        :return:
        """
        # set any property on a sliver that has a setter
        for k, v in kwargs.items():
            try:
                # we can set anything the sliver model has a setter for
                self.__getattribute__('set_' + k)(v)
            except AttributeError:
                raise RuntimeError(f'Unable to set property {k} on the sliver - no such property available')


class FimHelper:
    """
    Provides methods to load Graph Models and perform various operations on them
    """
    __neo4j_graph_importer = None

    @staticmethod
    def get_neo4j_importer(neo4j_config: dict = None) -> ABCGraphImporter:
        """
        get fim graph importer
        :return: Neo4jGraphImporter
        """
        logger = None
        if neo4j_config is None:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            neo4j_config = GlobalsSingleton.get().get_config().get_global_config().get_neo4j_config()
            logger = GlobalsSingleton.get().get_logger()

        if FimHelper.__neo4j_graph_importer is None:
            FimHelper.__neo4j_graph_importer = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                                                  pswd=neo4j_config["pass"],
                                                                  import_host_dir=neo4j_config["import_host_dir"],
                                                                  import_dir=neo4j_config["import_dir"], logger=logger)
        return FimHelper.__neo4j_graph_importer

    @staticmethod
    def get_networkx_importer(logger: logging.Logger = None) -> ABCGraphImporter:
        """
        get fim graph importer
        :return: Neo4jGraphImporter
        """
        if logger is None:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            logger = GlobalsSingleton.get().get_logger()

        return NetworkXGraphImporter(logger=logger)

    @staticmethod
    def get_arm_graph_from_file(*, filename: str, graph_id: str = None, neo4j_config: dict = None) -> ABCARMPropertyGraph:
        """
        Load specified file directly with no manipulations or validation
        :param filename:
        :param graph_id:
        :param neo4j_config neo4j_config
        :return:
        """
        neo4j_graph_importer = FimHelper.get_neo4j_importer(neo4j_config=neo4j_config)
        neo4_graph = neo4j_graph_importer.import_graph_from_file(graph_file=filename, graph_id=graph_id)
        site_arm = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=neo4_graph.graph_id,
                                                          importer=neo4j_graph_importer))

        site_arm.validate_graph()

        return site_arm

    @staticmethod
    def get_arm_graph(*, graph_id: str, neo4j_config: dict = None) -> ABCARMPropertyGraph:
        """
        Load arm graph from fim
        :param graph_id: graph_id
        :param neo4j_config neo4j_config
        :return: Neo4jARMGraph
        """
        neo4j_graph_importer = FimHelper.get_neo4j_importer(neo4j_config=neo4j_config)
        arm_graph = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=graph_id, importer=neo4j_graph_importer))
        if arm_graph.graph_exists():
            arm_graph.validate_graph()

        return arm_graph

    @staticmethod
    def get_experiment_topology(*, graph_id: str) -> ExperimentTopology:
        """
        Load Experiment Topology provide Graph Id
        :param graph_id Graph Id
        """
        graph = FimHelper.get_graph(graph_id=graph_id)
        asm_graph = Neo4jASMFactory.create(graph=graph)
        neo4j_topo = ExperimentTopology()
        neo4j_topo.cast(asm_graph=asm_graph)
        return neo4j_topo

    @staticmethod
    def get_graph(*, graph_id: str, neo4j_config: dict = None) -> ABCPropertyGraph:
        """
        Load arm graph from fim
        :param graph_id: graph_id.
        :param neo4j_config neo4j_config
        :return: Neo4jARMGraph
        """
        neo4j_graph_importer = FimHelper.get_neo4j_importer(neo4j_config=neo4j_config)
        arm_graph = Neo4jPropertyGraph(graph_id=graph_id, importer=neo4j_graph_importer)

        return arm_graph

    @staticmethod
    def get_neo4j_cbm_graph(graph_id: str, neo4j_config: dict = None) -> ABCCBMPropertyGraph:
        """
        Load cbm graph from fim
        :param graph_id: graph_id
        :param neo4j_config neo4j_config
        :return: Neo4jCBMGraph
        """
        neo4j_graph_importer = FimHelper.get_neo4j_importer(neo4j_config=neo4j_config)
        combined_broker_model = Neo4jCBMGraph(graph_id=graph_id,
                                              importer=neo4j_graph_importer,
                                              logger=neo4j_graph_importer.log)
        if combined_broker_model.graph_exists():
            combined_broker_model.validate_graph()
        return combined_broker_model

    @staticmethod
    def get_neo4j_cbm_graph_from_string_direct(*, graph_str: str, ignore_validation: bool = False) -> ABCCBMPropertyGraph:
        """
        Load Broker Query model graph from string
        :param graph_str: graph_str
        :param ignore_validation: ignore validation when set to true
        :return: Neo4jCBMGraph
        """
        neo4_graph = FimHelper.get_graph_from_string_direct(graph_str=graph_str)
        if neo4_graph.graph_exists() and not ignore_validation:
            neo4_graph.validate_graph()
        return Neo4jCBMFactory.create(neo4_graph)

    @staticmethod
    def get_graph_from_string_direct(*, graph_str: str, neo4j_config: dict = None) -> ABCPropertyGraph:
        """
        Load arm graph from fim
        :param graph_str: graph_str
        :param neo4j_config neo4j_config
        :return: Neo4jPropertyGraph
        """
        neo4j_graph_importer = FimHelper.get_neo4j_importer(neo4j_config=neo4j_config)
        graph = neo4j_graph_importer.import_graph_from_string_direct(graph_string=graph_str)

        return graph

    @staticmethod
    def get_graph_from_string(*, graph_str: str, neo4j_config: dict = None) -> ABCPropertyGraph:
        """
        Load arm graph from fim
        :param graph_str: graph_str
        :param neo4j_config neo4j_config
        :return: Neo4jPropertyGraph
        """
        neo4j_graph_importer = FimHelper.get_neo4j_importer(neo4j_config=neo4j_config)
        graph = neo4j_graph_importer.import_graph_from_string(graph_string=graph_str)

        return graph

    @staticmethod
    def get_networkx_graph_from_string(*, graph_str: str) -> ABCPropertyGraph:
        """
        Load arm graph from fim
        :param graph_str: graph_str
        :return: NetworkXPropertyGraph
        """
        networkx_graph_importer = FimHelper.get_networkx_importer()
        graph = networkx_graph_importer.import_graph_from_string(graph_string=graph_str)

        return graph

    @staticmethod
    def delete_graph(*, graph_id: str, neo4j_config: dict = None):
        """
        Delete a graph
        @param graph_id graph id
        @param neo4j_config neo4j_config
        """
        neo4j_graph_importer = FimHelper.get_neo4j_importer(neo4j_config=neo4j_config)
        neo4j_graph_importer.delete_graph(graph_id=graph_id)

    @staticmethod
    def get_delegation(delegated_capacities: Delegations, delegation_name: str) -> Capacities:
        """
        Get Delegated capacity given delegation name
        :param delegated_capacities: list of delegated capacities
        :param delegation_name: delegation name
        :return: capacity for specified delegation
        """
        delegation = delegated_capacities.get_by_delegation_id(delegation_name)
        return delegation.get_details() if delegation is not None else None

    @staticmethod
    def update_node(*, graph_id: str, sliver: BaseSliver):
        """
        Update Sliver Node in ASM
        :param graph_id:
        :param sliver:
        :return:
        """
        if sliver is None:
            return
        graph = FimHelper.get_graph(graph_id=graph_id)
        asm_graph = Neo4jASMFactory.create(graph=graph)
        neo4j_topo = ExperimentTopology()
        neo4j_topo.cast(asm_graph=asm_graph)

        node_name = sliver.get_name()
        if isinstance(sliver, NodeSliver) and node_name in neo4j_topo.nodes:
            node = neo4j_topo.nodes[node_name]
            node.set_properties(labels=sliver.labels,
                                label_allocations=sliver.label_allocations,
                                capacity_allocations=sliver.capacity_allocations,
                                reservation_info=sliver.reservation_info,
                                node_map=sliver.node_map,
                                management_ip=sliver.management_ip,
                                capacity_hints=sliver.capacity_hints)
            if sliver.attached_components_info is not None:
                graph_sliver = asm_graph.build_deep_node_sliver(node_id=sliver.node_id)
                diff = graph_sliver.diff(other_sliver=sliver)
                if diff is not None:
                    for cname in diff.removed.components:
                        reservation_info = ReservationInfo()
                        reservation_info.reservation_state = ReservationStates.Failed.name
                        node.components[cname].set_properties(reservation_info=reservation_info)

                topo_component_dict = node.components
                for component in sliver.attached_components_info.devices.values():
                    topo_component = topo_component_dict[component.get_name()]
                    topo_component.set_properties(labels=component.labels,
                                                  label_allocations=component.label_allocations,
                                                  capacity_allocations=component.capacity_allocations,
                                                  node_map=component.node_map)
                    # Update Mac address
                    if component.network_service_info is not None and \
                            component.network_service_info.network_services is not None:
                        topo_ifs_dict = topo_component.interfaces
                        for ns in component.network_service_info.network_services.values():
                            if ns.interface_info is None or ns.interface_info.interfaces is None:
                                continue
                            for ifs in ns.interface_info.interfaces.values():
                                topo_ifs = topo_ifs_dict[ifs.get_name()]
                                topo_ifs.set_properties(labels=ifs.labels,
                                                        label_allocations=ifs.label_allocations,
                                                        node_map=ifs.node_map)
                                if ifs.peer_labels is not None:
                                    topo_ifs.set_properties(peer_labels=ifs.peer_labels)

                                if ifs.capacities is not None:
                                    topo_ifs.set_properties(capacities=ifs.capacities)

        elif isinstance(sliver, NetworkServiceSliver) and node_name in neo4j_topo.network_services:
            node = neo4j_topo.network_services[node_name]
            node.set_properties(labels=sliver.labels,
                                label_allocations=sliver.label_allocations,
                                capacity_allocations=sliver.capacity_allocations,
                                reservation_info=sliver.reservation_info,
                                node_map=sliver.node_map,
                                gateway=sliver.gateway)
            if sliver.interface_info is not None:
                topo_ifs_dict = node.interfaces
                for ifs in sliver.interface_info.interfaces.values():
                    if ifs.get_name() not in node.interfaces:
                        continue
                    topo_ifs = topo_ifs_dict[ifs.get_name()]
                    topo_ifs.set_properties(labels=ifs.labels,
                                            label_allocations=ifs.label_allocations,
                                            node_map=ifs.node_map)

                    if ifs.peer_labels is not None:
                        topo_ifs.set_properties(peer_labels=ifs.peer_labels)

                    if ifs.capacities is not None:
                        topo_ifs.set_properties(capacities=ifs.capacities)

    @staticmethod
    def get_neo4j_asm_graph(*, slice_graph: str) -> ABCASMPropertyGraph:
        """
        Load Slice in Neo4j
        :param slice_graph: slice graph string
        :return: Neo4j ASM graph
        """
        neo4j_graph = FimHelper.get_graph_from_string(graph_str=slice_graph)
        asm = Neo4jASMFactory.create(graph=neo4j_graph)
        return asm

    @staticmethod
    def get_networkx_asm_graph(*, slice_graph: str) -> NetworkxASM:
        """
        Load Slice in NetworkX
        :param slice_graph: slice graph string
        :return: NetworkX ASM graph
        """
        networkx_graph = FimHelper.get_graph_from_string(graph_str=slice_graph)
        asm = NetworkXASMFactory.create(graph=networkx_graph)
        return asm

    @staticmethod
    def get_interface_sliver_mapping(ifs_node_id: str, slice_graph: ABCASMPropertyGraph) -> InterfaceSliverMapping:
        """
        Finds Peer Interface Sliver and parent information upto Network Node
        @param ifs_node_id node id of the Interface Sliver
        @param slice_graph Slice ASM
        @returns Interface Sliver Mapping
        """

        # Peer Connection point maps to Interface Sliver in ASM
        # This must always return only 1 IFS
        peer_interfaces = FimHelper.get_peer_interfaces(ifs_node_id=ifs_node_id, graph=slice_graph)
        if len(peer_interfaces) != 1:
            raise Exception(f"More than one Peer Interface Sliver found for IFS: {ifs_node_id}!")
        peer_ifs = next(iter(peer_interfaces))

        peer_ns_node_name, peer_ns_id = slice_graph.get_parent(node_id=peer_ifs.node_id,
                                                               rel=ABCPropertyGraph.REL_CONNECTS,
                                                               parent=ABCPropertyGraph.CLASS_NetworkService)

        component_name = None
        facility = False
        peer_site = None

        if peer_ifs.get_type() in [InterfaceType.DedicatedPort, InterfaceType.SharedPort]:
            component_name, component_id = slice_graph.get_parent(node_id=peer_ns_id, rel=ABCPropertyGraph.REL_HAS,
                                                                  parent=ABCPropertyGraph.CLASS_Component)

            node_name, node_id = slice_graph.get_parent(node_id=component_id, rel=ABCPropertyGraph.REL_HAS,
                                                        parent=ABCPropertyGraph.CLASS_NetworkNode)
        elif peer_ifs.get_type() == InterfaceType.FacilityPort:
            node_name, node_id = slice_graph.get_parent(node_id=peer_ns_id, rel=ABCPropertyGraph.REL_HAS,
                                                        parent=ABCPropertyGraph.CLASS_NetworkNode)
            node_sliver = slice_graph.build_deep_node_sliver(node_id=node_id)
            # Passing Facility Name instead of Node ID
            node_id = f"{node_sliver.get_site()},{node_name}"
            facility = True
        else:
            node_id = None
            peer_ns = slice_graph.build_deep_ns_sliver(node_id=peer_ns_id)

            # Peer Network Service is FABRIC L3VPN connected to a FABRIC Site
            # Determine the site to which AL2S Peered Interface is connected to
            for ifs in peer_ns.interface_info.interfaces.values():
                # Skip the peered interface
                if ifs.node_id == peer_ifs.node_id:
                    continue
                # Grab the first interface connected to a VM
                peer_nic_ifs_list = FimHelper.get_peer_interfaces(ifs_node_id=ifs.node_id,
                                                                  graph=slice_graph)
                ovs_ns_name, ovs_ns_id = slice_graph.get_parent(node_id=peer_nic_ifs_list[0].node_id,
                                                                rel=ABCPropertyGraph.REL_CONNECTS,
                                                                parent=ABCPropertyGraph.CLASS_NetworkService)
                ovs_ns = slice_graph.build_deep_ns_sliver(node_id=ovs_ns_id)

                peer_site = ovs_ns.get_site()

                if ovs_ns.get_type() != ServiceType.OVS:
                    # Peer node i.e. Facility Port
                    peer_node_name, peer_node_id = slice_graph.get_parent(node_id=ovs_ns.node_id,
                                                                          rel=ABCPropertyGraph.REL_HAS,
                                                                          parent=ABCPropertyGraph.CLASS_NetworkNode)

                    peer_node = slice_graph.build_deep_node_sliver(node_id=peer_node_id)

                    peer_site = f'{ovs_ns.get_site()},{peer_node.get_type()},{peer_node.get_name()}'
                break

        ret_val = InterfaceSliverMapping()
        ret_val.set_properties(peer_ifs=peer_ifs, peer_ns_id=peer_ns_id, component_name=component_name,
                               node_id=node_id, facility=facility, peer_site=peer_site)
        return ret_val

    @staticmethod
    def get_peer_interfaces(ifs_node_id: str, graph: ABCPropertyGraph,
                            interface_type: InterfaceType = None) -> List[InterfaceSliver]:
        """
        Finds Peer Interface Sliver and parent information upto Network Node
        @param ifs_node_id node id of the Interface Sliver
        @param graph Slice ASM
        @param interface_type Interface Type
        @returns Interface Sliver Mapping
        """
        result = []
        candidates = graph.find_peer_connection_points(node_id=ifs_node_id)
        if candidates is None:
            return result
        for c in candidates:
            clazzes, node_props = graph.get_node_properties(node_id=c)

            # Peer Connection point maps to Interface Sliver
            # Build Interface Sliver
            peer_ifs = ABCPropertyGraph.interface_sliver_from_graph_properties_dict(d=node_props)
            if interface_type is not None and peer_ifs.get_type() == interface_type:
                result.append(peer_ifs)
                break
            else:
                result.append(peer_ifs)
        return result

    @staticmethod
    def get_site_interface_sliver(*, component: ComponentSliver or NodeSliver, local_name: str,
                                  region: str = None, device_name: str = None) -> InterfaceSliver or None:
        """
        Get Interface Sliver (child of Component Sliver) with a local name

        E.g: Component renc-w3-nic2 has two connection points; this function returns the connection point whose
        name matches the local name
        renc-w3         => renc-w3-nic2         => renc-w3-nic2-l2ovs   => renc-w3-nic2-p1
        [Network Node]     [Component]             [Network Service]       [Connection Point]
                                                                        => renc-w3-nic2-p2
                                                                           [Connection Point]

        @param component Component Sliver
        @param local_name Local Name
        @param region region
        @param device_name device name
        @return Interface sliver
        """
        for ns in component.network_service_info.network_services.values():
            # Filter on region
            if region is not None:
                result = list(filter(lambda x: (region in x.labels.region), ns.interface_info.interfaces.values()))
            else:
                result = list(ns.interface_info.interfaces.values())

            # Filter on device name
            if device_name is not None:
                result = list(filter(lambda x: (device_name in x.labels.device_name), result))

            if local_name is not None:
                result = list(filter(lambda x: (local_name in x.labels.local_name), result))

            return random.choice(result)
            '''
            # Return specific interface where local name matches
            if local_name is not None:
                return next(x for x in ns.interface_info.interfaces.values() if local_name in x.get_name())
            else:
                # Return an interface chosen randomly
                if region is None:
                    return random.choice(list(ns.interface_info.interfaces.values()))
                # Return an interface chosen randomly for a specific region
                else:
                    result = list(filter(lambda x: (region in x.labels.region), ns.interface_info.interfaces.values()))
                    return random.choice(result)
            '''
        return None

    @staticmethod
    def get_owners(*, bqm: ABCCBMPropertyGraph, node_id: str,
                   ns_type: ServiceType) -> Tuple[NodeSliver, NetworkServiceSliver, NetworkServiceSliver]:
        """
        Get owner switch and network service of a Connection Point from BQM
        @param bqm BQM graph
        @param node_id Connection Point Node Id
        @param ns_type Network Service Type
        @return Owner Switch and Network Service
        """
        mpls_ns_name, mpls_ns_id = bqm.get_parent(node_id=node_id, rel=ABCPropertyGraph.REL_CONNECTS,
                                                  parent=ABCPropertyGraph.CLASS_NetworkService)

        mpls_ns = bqm.build_deep_ns_sliver(node_id=mpls_ns_id)

        sw_name, sw_id = bqm.get_parent(node_id=mpls_ns_id, rel=ABCPropertyGraph.REL_HAS,
                                        parent=ABCPropertyGraph.CLASS_NetworkNode)

        switch = bqm.build_deep_node_sliver(node_id=sw_id)

        requested_ns = mpls_ns
        if ns_type in Constants.L3_SERVICES:
            for ns in switch.network_service_info.network_services.values():
                if ns_type == ns.get_type():
                    requested_ns = ns
                    break

        return switch, mpls_ns, requested_ns

    @staticmethod
    def get_parent_node(*, graph_model: ABCPropertyGraph, component: Component = None, interface: Interface = None,
                        sliver: bool = True) -> Tuple[Union[NodeSliver, NetworkServiceSliver, None], str]:
        node = None
        if component is not None:
            node_name, node_id = graph_model.get_parent(node_id=component.node_id, rel=ABCPropertyGraph.REL_HAS,
                                                        parent=ABCPropertyGraph.CLASS_NetworkNode)
            if sliver:
                node = graph_model.build_deep_node_sliver(node_id=node_id)
        elif interface is not None:
            node_name, node_id = graph_model.get_parent(node_id=interface.node_id, rel=ABCPropertyGraph.REL_CONNECTS,
                                                        parent=ABCPropertyGraph.CLASS_NetworkService)
            if sliver:
                node = graph_model.build_deep_ns_sliver(node_id=node_id)
        else:
            raise Exception("Invalid Arguments - component/interface both are None")

        return node, node_id

    @staticmethod
    def prune_graph(*, graph_id: str) -> ExperimentTopology:
        """
        Load arm graph from fim and prune all nodes with reservation_state = reservation_state
        :param graph_id: graph_id
        :return: ExperimentTopology
        """
        slice_topology = FimHelper.get_experiment_topology(graph_id=graph_id)
        slice_topology.prune(reservation_state=ReservationStates.Failed.name)
        slice_topology.prune(reservation_state=ReservationStates.Closed.name)

        return slice_topology



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List

from fim.graph.abc_property_graph import ABCPropertyGraph
from fabric_cf.actor.core.apis.abc_mgmt_actor import ABCMgmtActor

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
    from fabric_mb.message_bus.messages.slice_avro import SliceAvro
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.manage.messages.client_mng import ClientMng
    from fabric_cf.actor.security.auth_token import AuthToken


class ABCMgmtServerActor(ABCMgmtActor):
    """
    Interface for Management Server Actor
    """
    @abstractmethod
    def get_broker_reservations(self, *, id_token: str = None) -> List[ReservationMng]:
        """
        Returns all reservations for which this actor acts as a broker.
        @param id_token id token
        @return list of the reservations
        """

    @abstractmethod
    def get_inventory_slices(self, *, id_token: str = None) -> List[SliceAvro]:
        """
        Obtains all slices holding inventory, i.e., resources that can be
        delegated to other actors.
        @param id_token id token
        @return list of slices
        """

    @abstractmethod
    def get_inventory_reservations(self, *, slice_id: ID = None, id_token: str = None) -> List[ReservationMng]:
        """
        Returns all resources held by this actor that can be used for delegations
        to client actors.
        @param slice_id slice id
        @param id_token id token
        @return list of reservations
        """

    @abstractmethod
    def get_client_slices(self, *, id_token: str = None) -> List[SliceAvro]:
        """
        Obtains all slices that hold delegated resources to other actors.
        @param id_token id token
        @return list of client slices
        """

    @abstractmethod
    def add_client_slice(self, *, slice_mng: SliceAvro) -> ID:
        """
        Adds a new client slice.
        @param slice_mng slice to be added
        @return sliceid of the added slice
        """

    @abstractmethod
    def get_clients(self, *, guid: ID = None, id_token: str = None) -> List[ClientMng]:
        """
        Returns all registered clients of this server actor.
        @param guid client guid
        @param id_token id token
        @return list of clients
        """

    @abstractmethod
    def register_client(self, *, client: ClientMng, kafka_topic: str) -> bool:
        """
        Registers a new client
        @param client client
        @param kafka_topic Kafka topic
        @return true for success; false otherwise
        """

    @abstractmethod
    def update_client(self, *, client: ClientMng, kafka_topic: str) -> bool:
        """
        Update a client
        @param client client
        @param kafka_topic Kafka topic
        @return true for success; false otherwise
        """

    @abstractmethod
    def unregister_client(self, *, guid: ID) -> bool:
        """
        Unregisters the specified client.
        @param guid client guid
        @return true for success; false otherwise
        """

    @abstractmethod
    def get_client_reservations(self, *, slice_id: ID = None, id_token: str = None) -> List[ReservationMng]:
        """
        Obtains all client reservations.
        @param slice_id slice id
        @param id_token id token
        @return list of client reservations
        """

    @abstractmethod
    def advertise_resources(self, *, delegation: ABCPropertyGraph, delegation_name: str, client: AuthToken) -> ID:
        """
        Advertise resources to the broker
        @param delegation: delegation
        @param delegation_name delegation name
        @param client: client
        @return return the reservation id
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_delegation import DelegationState
from fabric_cf.actor.core.apis.abc_timer_task import ABCTimerTask

if TYPE_CHECKING:
    from fabric_cf.actor.core.kernel.rpc_request import RPCRequest


class ClaimTimeout(ABCTimerTask):
    """
    Claim Timeout
    """
    def __init__(self, *, req: RPCRequest):
        self.req = req

    def execute(self):
        """
        Process a claim timeout
        """
        self.req.actor.get_logger().debug("Claim timeout. Delegation= {}".format(self.req.delegation))
        self.req.actor.get_logger().error("Failing delegation {} due to expired claim timeout".format(
            self.req.delegation))
        self.req.actor.fail_delegation(did=self.req.get_delegation().get_delegation_id(),
                                       message="Timeout during claim. Please remove the delegation and retry later")


class ReclaimTimeout(ABCTimerTask):
    """
    Reclaim timeout
    """
    def __init__(self, *, req: RPCRequest):
        self.req = req

    def execute(self):
        """
        Process a reclaim timeout
        """
        self.req.actor.get_logger().debug("Reclaim timeout. delegation= {}".format(self.req.delegation))
        if self.req.delegation.get_state() == DelegationState.Delegated:
            self.req.actor.get_logger().error("Failing delegation {} due to expired reclaim timeout".format(
                self.req.reservation))
            self.req.actor.fail_delegation(did=self.req.get_delegation().get_delegation_id(),
                                           message="Timeout during claim. Please remove the delegation and retry later")
        else:
            self.req.actor.get_logger().debug("Reclaim has already completed")



import connexion

from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response import slivers_controller as rc


def slivers_get(slice_id, as_self=None):  # noqa: E501
    """Retrieve a listing of user slivers

    Retrieve a listing of user slivers # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str
    :param as_self: GET object as Self
    :type as_self: bool

    :rtype: Slivers
    """
    return rc.slivers_get(slice_id=slice_id, as_self=as_self)


def slivers_sliver_id_get(slice_id, sliver_id, as_self=None):  # noqa: E501
    """slivers properties

    Retrieve Sliver properties # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str
    :param sliver_id: Sliver identified by universally unique identifier
    :type sliver_id: str
    :param as_self: GET object as Self
    :type as_self: bool

    :rtype: Slivers
    """
    return rc.slivers_sliver_id_get(slice_id=slice_id, sliver_id=sliver_id, as_self=as_self)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author Komal Thareja (kthare10@renci.org)
from bisect import bisect_left

from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver
from fim.user import ComponentType
from fim.user.topology import TopologyDiff, TopologyDiffTuple

from fabric_cf.actor.security.pdp_auth import ActionId


def binary_search(*, a, x):
    """
    Binary search element x in list a
    """
    i = bisect_left(a, x)
    if i != len(a) and a[i] == x:
        return i
    else:
        return -1


def sliver_to_str(*, sliver: BaseSliver):
    if isinstance(sliver, NodeSliver):
        return node_sliver_to_str(sliver=sliver)

    if isinstance(sliver, NetworkServiceSliver):
        return ns_sliver_to_str(sliver=sliver)


def node_sliver_to_str(*, sliver: NodeSliver):
    result = str(sliver)
    nic_types = [ComponentType.SharedNIC, ComponentType.SmartNIC]
    if sliver.attached_components_info is not None:
        for c in sliver.attached_components_info.devices.values():
            result += f"\nComponent: {c}"
            if c.get_type() not in nic_types:
                continue
            if c.network_service_info is not None and c.network_service_info.network_services is not None:
                for ns in c.network_service_info.network_services.values():
                    result += f"\nNS: {ns}"
                    if ns.interface_info is not None and ns.interface_info.interfaces is not None:
                        for i in ns.interface_info.interfaces.values():
                            result += f"\nIFS: {i}"
    return result


def ns_sliver_to_str(*, sliver: NetworkServiceSliver):
    result = str(sliver)
    for interface in sliver.interface_info.interfaces.values():
        result += f"\nIFS: {interface}"
    return result


def translate_avro_message_type_pdp_action_id(*, message_name: str) -> ActionId:
    if message_name == AbcMessageAvro.claim_resources:
        return ActionId.claim
    elif message_name == AbcMessageAvro.reclaim_resources:
        return ActionId.reclaim
    elif message_name == AbcMessageAvro.get_slices_request or message_name == AbcMessageAvro.get_reservations_request \
        or message_name == AbcMessageAvro.get_reservations_state_request or \
        message_name == AbcMessageAvro.get_delegations or message_name == AbcMessageAvro.get_reservation_units_request \
            or message_name == AbcMessageAvro.get_unit_request or message_name == AbcMessageAvro.get_broker_query_model_request:
        return ActionId.query
    elif message_name == AbcMessageAvro.remove_slice:
        return ActionId.delete
    elif message_name == AbcMessageAvro.close_reservations:
        return ActionId.close
    elif message_name == AbcMessageAvro.remove_reservation:
        return ActionId.delete
    elif message_name == AbcMessageAvro.extend_reservation:
        return ActionId.renew
    else:
        return ActionId.noop



import connexion

from fabric_cf.orchestrator.swagger_server.models import SlicesPost
from fabric_cf.orchestrator.swagger_server.models.slice_details import SliceDetails  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slices import Slices  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slices_post import SlicesPost  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status200_ok_no_content import Status200OkNoContent  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response import slices_controller as rc


def slices_create_post(body, name, ssh_key, lease_end_time=None):  # noqa: E501
    """Create slice

    Request to create slice as described in the request. Request would be a graph ML describing the requested resources.
    Resources may be requested to be created now or in future. On success, one or more slivers are allocated, containing
    resources satisfying the request, and assigned to the given slice. This API returns list and description of the
    resources reserved for the slice in the form of Graph ML. Orchestrator would also trigger provisioning of these
    resources asynchronously on the appropriate sites either now or in the future as requested. Experimenter can
    invoke get slice API to get the latest state of the requested resources.   # noqa: E501

    :param body: 
    :type body: dict | bytes
    :param name: Slice Name
    :type name: str
    :param ssh_key: User SSH Key
    :type ssh_key: str
    :param lease_end_time: Lease End Time for the Slice
    :type lease_end_time: str

    :rtype: Slivers
    """
    post_body = SlicesPost()
    post_body.graph_model = body.decode("utf-8")
    post_body.ssh_keys = [ssh_key]
    return rc.slices_create_post(post_body, name, lease_end_time)


def slices_creates_post(body, name, lease_end_time=None):  # noqa: E501
    """Create slice

    Request to create slice as described in the request. Request would be a graph ML describing the requested resources.
    Resources may be requested to be created now or in future. On success, one or more slivers are allocated,
    containing resources satisfying the request, and assigned to the given slice. This API returns list and
    description of the resources reserved for the slice in the form of Graph ML. Orchestrator would also trigger
    provisioning of these resources asynchronously on the appropriate sites either now or in the future as requested.
    Experimenter can invoke get slice API to get the latest state of the requested resources.   # noqa: E501

    :param body: Create new Slice
    :type body: dict | bytes
    :param name: Slice Name
    :type name: str
    :param lease_end_time: Lease End Time for the Slice
    :type lease_end_time: str

    :rtype: Slivers
    """
    if connexion.request.is_json:
        body = SlicesPost.from_dict(connexion.request.get_json())  # noqa: E501
    return rc.slices_create_post(body, name, lease_end_time)


def slices_delete_delete():  # noqa: E501
    """Delete all slices for a User within a project.

    Delete all slices for a User within a project. User identity email and project id is available in the
    bearer token.  # noqa: E501


    :rtype: Status200OkNoContent
    """
    return rc.slices_delete_delete()


def slices_delete_slice_id_delete(slice_id):  # noqa: E501
    """Delete slice.

    Request to delete slice. On success, resources associated with slice or sliver are stopped if necessary,
    de-provisioned and un-allocated at the respective sites.  # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str

    :rtype: Status200OkNoContent
    """
    return rc.slices_delete_slice_id_delete(slice_id)


def slices_get(name=None, as_self=None, states=None, limit=None, offset=None):  # noqa: E501
    """Retrieve a listing of user slices

    Retrieve a listing of user slices. It returns list of all slices belonging to all members in a project when
    &#x27;as_self&#x27; is False otherwise returns only the all user&#x27;s slices in a project. # noqa: E501

    :param name: Search for Slices with the name
    :type name: str
    :param as_self: GET object as Self
    :type as_self: bool
    :param states: Search for Slices in the specified states
    :type states: List[str]
    :param limit: maximum number of results to return per page (1 or more)
    :type limit: int
    :param offset: number of items to skip before starting to collect the result set
    :type offset: int

    :rtype: Slices
    """
    return rc.slices_get(name, states, limit, offset, as_self=as_self)


def slices_modify_slice_id_accept_post(slice_id):  # noqa: E501
    """Accept the last modify an existing slice

    Accept the last modify and prune any failed resources from the Slice. Also return the accepted slice
    model back to the user.   # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str

    :rtype: SliceDetails
    """
    return rc.slices_modify_slice_id_accept_post(slice_id)


def slices_modify_slice_id_put(body, slice_id):  # noqa: E501
    """Modify an existing slice

    Request to modify an existing slice as described in the request. Request would be a graph ML describing the
    experiment topolgy expected after a modify. The supported modify actions include adding or removing nodes,
    components, network services or interfaces of the slice. On success, one or more slivers are allocated,
    containing resources satisfying the request, and assigned to the given slice. This API returns list and
    description of the resources reserved for the slice in the form of Graph ML. Orchestrator would also trigger
    provisioning of these resources asynchronously on the appropriate sites either now or in the future as requested.
    Experimenter can invoke get slice API to get the latest state of the requested resources.   # noqa: E501

    :param body: Modify a Slice
    :type body: dict | bytes
    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str

    :rtype: Slivers
    """
    return rc.slices_modify_slice_id_put(body, slice_id)


def slices_renew_slice_id_post(slice_id, lease_end_time):  # noqa: E501
    """Renew slice

    Request to extend slice be renewed with their expiration extended. If possible, the orchestrator should extend the
    slivers to the requested expiration time, or to a sooner time if policy limits apply.  # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str
    :param lease_end_time: New Lease End Time for the Slice
    :type lease_end_time: str

    :rtype: Status200OkNoContent
    """
    return rc.slices_renew_slice_id_post(slice_id, lease_end_time)


def slices_slice_id_get(slice_id, graph_format, as_self=None):  # noqa: E501
    """slice properties

    Retrieve Slice properties # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str
    :param graph_format: graph format
    :type graph_format: str
    :param as_self: GET object as Self
    :type as_self: bool

    :rtype: SliceDetails
    """
    return rc.slices_slice_id_get(slice_id, graph_format, as_self=as_self)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import os
import threading
from logging.handlers import RotatingFileHandler


class LogHelper:
    @staticmethod
    def make_logger(*, log_dir: str, log_file: str, log_level, log_retain: int, log_size: int, logger: str,
                    log_format: str = None):
        """
        Detects the path and level for the log file from the actor config and sets
        up a logger. Instead of detecting the path and/or level from the
        config, a custom path and/or level for the log file can be passed as
        optional arguments.

       :param log_dir: Log directory
       :param log_file
       :param log_level
       :param log_retain
       :param log_size
       :param logger
       :param log_format
       :return: logging.Logger object
        """
        log_path = f"{log_dir}/{log_file}"

        if log_path is None:
            raise RuntimeError('The log file path must be specified in config or passed as an argument')

        if log_level is None:
            log_level = logging.INFO

        # Set up the root logger
        log = logging.getLogger(logger)
        log.setLevel(log_level)
        default_log_format = \
            '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s-%(thread_id)s]- %(levelname)s - %(message)s'
        if log_format is not None:
            default_log_format = log_format

        os.makedirs(os.path.dirname(log_path), exist_ok=True)

        file_handler = RotatingFileHandler(log_path, backupCount=int(log_retain), maxBytes=int(log_size))
        file_handler.addFilter(LogHelper.thread_id_filter)
        file_handler.setFormatter(logging.Formatter(default_log_format))
        log.addHandler(file_handler)

        #logging.basicConfig(handlers=[file_handler], format=log_format, force=True)
        #file_handler.addFilter(LogHelper.thread_id_filter)
        return log

    @staticmethod
    def thread_id_filter(record):
        """Inject thread_id to log records"""
        record.thread_id = threading.get_native_id()
        return record



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime

from fim.slivers.base_sliver import BaseSliver

from fabric_cf.actor.core.apis.abc_base_plugin import ABCBasePlugin
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import ResourcesException


if TYPE_CHECKING:
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.util.notice import Notice
    from fabric_cf.actor.core.apis.abc_concrete_set import ABCConcreteSet
    from fabric_cf.actor.core.util.resource_type import ResourceType
    from fabric_cf.actor.core.kernel.poa import Poa


class ResourceSet:
    """
    ResourceSet is an abstract set of resources describing some number of
    resource units of a given type, e.g., to represent a resource request. A
    ResourceSet with an attached ConcreteSet is "concrete." The ConcreteSet binds
    real resources (or a promise) for some or all of the abstract units in the
    set: for example, a concrete ResourceSet can represent a ticket or
    lease.
    """
    def __init__(self, *, concrete: ABCConcreteSet = None, gained: ABCConcreteSet = None,
                 lost: ABCConcreteSet = None, modified: ABCConcreteSet = None,
                 rtype: ResourceType = None, sliver: BaseSliver = None, units: int = None):
        # What type of resources does this set contain. The meaning/assignment of
        # type values is an externally defined convention of interacting actors.
        self.type = rtype
        # How many units (abstract) the set contains. This count reflects the
        # resources intended or requested for this set. For an active reservation
        # in steady state, the unit count will typically match the number of
        # concrete resources, but it might not match if the resource set is in flux
        # for some reason. For an inventory set, the abstract count reflects the
        # original size of the inventory, independent of any allocations extracted
        # from it.
        self.units = 0
        # Concrete resources.
        self.resources = concrete
        if concrete is not None:
            self.units = concrete.get_units()
        else:
            self.units = units
        # Sliver
        self.sliver = sliver
        # The previous value of the sliver. This is essential for
        # supporting recovery on Authority.
        self.previous_sliver = None
        # A set of resources recently ejected from the resource set, pending
        # processing, e.g., by a "probe" method.
        self.released = None
        # A recent update to the concrete resource set, pending processing by a
        # "service" method. Client-side only (i.e., for ticket or lease updates
        # through callback interface).
        self.updated = None
        # Recent additions of resources to the concrete set, pending processing by
        # a "service" method. Only for authority role only.
        self.gained = gained
        # Recently lost resources pending processing by a "service" method. Only
        # for authority role only.
        self.lost = lost
        # Recently changed resources pending a processing by a "service" method.
        # set.
        self.modified = modified
        # Reservation with which this set is associated.
        self.rid = None
        self.is_closing = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['released']
        del state['updated']
        del state['gained']
        del state['lost']
        del state['modified']
        del state['rid']
        del state['is_closing']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.released = None
        self.updated = None
        self.gained = None
        self.lost = None
        self.modified = None
        self.rid = None
        self.is_closing = False

    def restore(self, *, plugin: ABCBasePlugin, reservation: ABCReservationMixin):
        """
        Restore post stateful restart
        @param plugin plugin
        @param reservation reservation
        """
        if reservation is not None:
            self.rid = reservation.get_reservation_id()

        if self.resources is not None:
            self.resources.restore(plugin=plugin, reservation=reservation)

    def abstract_clone(self):
        """
        Clones the set, but without any of the concrete sets. Used on Authority
        and Service Manager to create a ResourceSet to hold resources, given a
        ResourceSet holding a ticket.
        @return a resources set that is a copy of the current but without any concrete sets.
        """
        result = ResourceSet(units=self.units, rtype=self.type, sliver=self.sliver)
        return result

    def collect_released(self) -> ResourceSet:
        """
        Any units that fail or are rejected/released asynchronously accumulate
        within the ConcreteSet until collected. These are cached by a
        prepareProbe.
        @returns a ResourceSet
        @raises Exception in case of error
        """
        result = None
        if self.released is not None:
            result = ResourceSet(concrete=self.released, rtype=self.type)
            self.released = None
        return result

    def merge_properties(self, *, reservation: ABCReservationMixin, resource_set):
        """
        Merge properties
        @param reservation reservation
        @param resource_set resource set
        """
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.INVALID_ARGUMENT)

        self.sliver = resource_set.get_sliver()

    def delta_update(self, *, reservation: ABCReservationMixin, resource_set):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.INVALID_ARGUMENT)

        if self.resources is None:
            # in case of close for a canceled reservation.
            if resource_set.gained is None:
                return
            # first time we give concrete resources to this resource set
            # Since this set has no concrete resources, we can only gain
            # resources. Lost and modified have no meaning in this case. Assert just in case
            if resource_set.lost is not None or resource_set.modified is not None:
                return
            # take the units and type
            self.units = resource_set.gained.get_units()
            self.type = resource_set.type
            self.resources = resource_set.gained.clone_empty()
            self.resources.setup(reservation=reservation)
            self.merge_properties(reservation=reservation, resource_set=resource_set)
            self.gained = resource_set.gained
        else:
            self.type = resource_set.type
            difference = 0

            if resource_set.lost is not None:
                self.lost = resource_set.lost
                difference -= resource_set.lost.get_units()

            if resource_set.modified is not None:
                self.modified = resource_set.modified

            self.units += difference
            self.previous_sliver = self.sliver
            self.merge_properties(reservation=reservation, resource_set=resource_set)

    def fix_abstract_units(self):
        """
        Sets the number of abstract units to equal the number of concrete units.
        """
        if self.resources is None:
            self.units = self.resources.get_units()
        else:
            self.units = 0

    def full_update(self, *, reservation: ABCReservationMixin, resource_set):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.INVALID_ARGUMENT)

        # take the units and the type
        self.units = resource_set.units
        self.type = resource_set.type
        # take in the sliver
        self.previous_sliver = self.sliver
        self.merge_properties(reservation=reservation, resource_set=resource_set)

        # make a concrete set if the current concrete set is None
        if self.resources is None:
            self.resources = resource_set.resources.clone_empty()
            self.resources.setup(reservation=reservation)
        # remember the update so that it can be processed later
        self.updated = resource_set.resources

    def get_concrete_units(self, *, when: datetime = None) -> int:
        """
        Estimate the concrete resource units the resource set will contain at the
        specified date.
        @params when: the date
        @returns number of concrete units
        """
        if self.resources is None:
            return 0
        elif when is None:
            return self.resources.get_units()
        else:
            return self.resources.holding(when=when)

    def get_deficit(self) -> int:
        """
        Returns the number of concrete units needed or in excess in this resource
        set.
        @returns number of units in excess or needed
        """
        result = self.units
        if self.resources is not None:
            result -= self.resources.get_units()

        return result

    def get_poa_info(self) -> dict:
        return self.resources.get_poa_info()

    def get_notices(self) -> Notice:
        """
        Returns a string of notices or events pertaining to the underlying
        resources. The event notices are consumed: subsequent calls return only
        new information. May return null.
        @returns Notice
        """
        if self.resources is None:
            return None
        return self.resources.get_notices()

    def get_sliver(self) -> BaseSliver:
        """
        Return sliver
        @param sliver
        """
        return self.sliver

    def get_reservation_id(self) -> ID:
        """
        Returns the reservation identifier attached to this resource set.
        @returns reservation identifier
        """
        return self.rid

    def get_resources(self) -> ABCConcreteSet:
        """
        Returns the concrete resources.
        @returns concrete resource set
        """
        return self.resources

    def get_site_proxy(self):
        """
        Returns a proxy to the site authority, which owns the resources
        represented in the set.
        @returns site authority proxy.
        @raises Exception in case of error
        """
        if self.resources is not None:
            return self.resources.get_site_proxy()
        return None

    def get_type(self) -> ResourceType:
        """
        Returns the resource type of the set.
        @returns resource type
        """
        return self.type

    def get_units(self) -> int:
        """
        Returns the number of abstract units in the set.
        @returns number of abstract units
        """
        return self.units

    def is_active(self) -> bool:
        """
        Checks if the resource set is active: allocated units are active.
        @returns true if this ResourceSet is active
        """
        if self.resources is not None:
            return self.resources.is_active()
        return False

    def is_closed(self) -> bool:
        """
        Checks if the resource set is closed: there are no active units. Do not
        call this method unless the set had a close in progress: a set with
        failed units or one that has not yet been activated may register as
        "closed".
        @returns true if this ResourceSet is active
        """
        if self.resources is None:
            return True
        if self.resources.get_units() == 0:
            return True
        return False

    def is_empty(self) -> bool:
        if self.updated is not None and self.updated.get_units() > 0:
            return False
        if self.gained is not None and self.gained.get_units() > 0:
            return False
        if self.lost is not None and self.lost.get_units() > 0:
            return False
        if self.modified is not None and self.modified.get_units() > 0:
            return False
        return True

    def prepare_probe(self):
        """
        Prepares a probe: updates ConcreteSet to reflect underlying resource
        status.
        @raises Exception in case of error
        """
        if self.resources is not None:
            self.resources.probe()
            if self.released is None:
                self.released = self.resources.collect_released()

    def probe(self):
        return

    def service_check(self):
        if self.resources is None:
            raise ResourcesException("Internal Error: WARNING: service post-op call on non-concrete reservation")

    def close(self):
        """
        Initiate close on the concrete resources
        """
        if not self.is_closing:
            self.is_closing = True
            self.resources.close()

    def service_extend(self):
        """
        Complete service for a term extension (server side).
        """
        self.service_check()
        # An elastic reservation can change concrete resources on extend. The
        # modifications are left in update/gained/lost by *Change() above. On
        # agent the concrete is updated synchronously in SoftChange, so this
        # code segment applies to authority only.
        my_gained = self.gained
        self.gained = None
        # An elastic reservation can change concrete resources on extend. The
        # modifications are left in update/gained/lost by *Change() above. On
        # agent the concrete is updated synchronously in SoftChange, so this
        # code segment applies to authority only.
        my_lost = self.lost
        self.lost = None
        # An elastic reservation can change concrete resources on extend. The
        # modifications are left in update/gained/lost by *Change() above. On
        # agent the concrete is updated synchronously in SoftChange, so this
        # code segment applies to authority only.
        my_modified = self.modified
        self.modified = None
        if my_gained is not None:
            self.resources.add(concrete_set=my_gained, configure=True)
        if my_lost is not None:
            self.resources.remove(concrete_set=my_lost, configure=True)
        if my_modified is not None:
            self.resources.modify(sliver=self.sliver)

    def service_modify(self):
        """
        Complete service for a term extension (server side).
        @raises Exception in case of error
        """
        self.service_check()
        self.resources.modify(sliver=self.sliver)

    def service_reserve_site(self):
        cs = None
        if self.gained is not None:
            cs = self.gained
            self.gained = None

        if cs is not None:
            self.resources.add(concrete_set=cs, configure=True)

    def service_update(self, *, reservation: ABCReservationMixin):
        """
        Service a resource set update. Any changes to existing
        concrete resources should have been left in "updated" by an update
        operation.
        @params reservation: reservation
        @raises Exception in case of error
        """
        cs = None
        if self.updated is not None:
            cs = self.updated
            self.updated = None
        if cs is not None:
            self.resources.change(concrete_set=cs, configure=True)

    def set_sliver(self, *, sliver: BaseSliver):
        """
        Sets the request sliver.
        @params sliver : request sliver
        """
        self.sliver = sliver

    def set_reservation_id(self, *, rid: ID):
        """
        Attaches the reservation identifier to the set.
        @params rid reservation identifier
        """
        self.rid = rid

    def set_resources(self, *, cset: ABCConcreteSet):
        """
        Set the concrete resources. Used by proxies.
        @params cset :concrete resource set
        """
        self.resources = cset

    def set_type(self, *, rtype: ResourceType):
        """
        Sets the resource type for the set.
        @params rtype : resource type
        """
        self.type = rtype

    def set_units(self, *, units: int):
        """
        Sets the number of abstract units in the set.
        @params units: number of abstract units
        """
        self.units = units

    def setup(self, *, reservation: ABCReservationMixin):
        """
        Passes information about the containing reservation to the concrete set.
        @params reservation: containing reservation
        """
        if self.resources is not None:
            self.resources.setup(reservation=reservation)

    def __str__(self):
        result = f"rset: units=[{self.units}] "
        if self.resources is not None:
            result += f" concrete:[{self.resources}]"
        return result

    def update(self, *, reservation: ABCReservationMixin, resource_set: ResourceSet):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.INVALID_ARGUMENT)

        if resource_set.resources is not None:
            self.full_update(reservation=reservation, resource_set=resource_set)
        else:
            self.delta_update(reservation=reservation, resource_set=resource_set)

    def update_properties(self, *, reservation: ABCReservationMixin, resource_set):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.INVALID_ARGUMENT)

        self.merge_properties(reservation=reservation, resource_set=resource_set)

    def validate(self):
        """
        Validates a fresh ResourceSet passed in from outside
        @raises Exception in case of error thrown if the set is determined to be invalid
        """
        if self.units < 0:
            raise ResourcesException("invalid unit count:{}".format(self.units))

    def validate_incoming(self):
        """
        Validates a ResourceSet in an incoming ticket or lease
        request (server) or in an incoming ticket or lease update (client).
        Called for each incoming request/update to check validity with no locks
        held.
        @raises Exception in case of error
        """
        self.validate()
        if self.resources is not None:
            self.resources.validate_incoming()

    def validate_incoming_ticket(self, *, term: Term):
        """
        Validate match between abstract and concrete ResourceSet in a ResourceSet
        representing an incoming ticket.
        @params t : optional term associated with ResourceSet
        @raises Exception in case of error if validation fails
        """
        if self.resources is None:
            if self.units != 0:
                raise ResourcesException("no resources to back incoming ticket")
            return
        if self.resources.get_units() != self.units:
            raise ResourcesException("size mismatch on incoming ticket {} != {}".format(self.resources.get_units(),
                                                                                        self.units))
        self.resources.validate_concrete(rtype=self.type, units=self.units, term=term)

    def validate_outgoing(self):
        """
        Validates a ResourceSet that is about to be sent to another
        actor. Client-side only.
        @raises Exception in case of error
        """
        self.validate()
        if self.resources is not None:
            self.resources.validate_outgoing()

    def clone(self):
        clone = ResourceSet(units=self.units, rtype=self.type, sliver=self.sliver)
        clone.resources = self.resources.clone()
        return clone

    def service_poa(self, *, poa: Poa):
        """
        Complete service for a POA.
        """
        self.service_check()
        self.resources.poa(poa=poa)



import unittest

from fim.graph.abc_property_graph import ABCPropertyGraph, GraphFormat
from fim.graph.neo4j_property_graph import Neo4jGraphImporter, Neo4jPropertyGraph
from fim.graph.resources.neo4j_arm import Neo4jARMGraph
from fim.graph.resources.neo4j_cbm import Neo4jCBMFactory, Neo4jCBMGraph
from fabric_cf.actor.fim.plugins.broker.aggregate_bqm_plugin import AggregatedBQMPlugin

"""
Test of an ABQM plugin
"""
import yaml
neo4j = None
with open("./config/config.test.yaml", 'r') as stream:
    try:
        config_dict = yaml.safe_load(stream)
        neo4j = config_dict["neo4j"]
    except yaml.YAMLError as exc:
        print(exc)


class ABQM_Test(unittest.TestCase):
    n4j_imp = Neo4jGraphImporter(url=neo4j["url"], user=neo4j["user"],
                                 pswd=neo4j["pass"],
                                 import_host_dir=neo4j["import_host_dir"],
                                 import_dir=neo4j["import_dir"])

    def test_abqm(self):
        self.n4j_imp.delete_all_graphs()
        # these are produced by substrate tests
        site_ads = ['../../../neo4j/Network-ad.graphml', '../../../neo4j/LBNL-ad.graphml',
                    '../../../neo4j/RENCI-ad.graphml',
                    '../../../neo4j/UKY-ad.graphml']

        cbm = Neo4jCBMGraph(importer=self.n4j_imp)

        adm_ids = dict()

        for ad in site_ads:
            plain_neo4j = self.n4j_imp.import_graph_from_file_direct(graph_file=ad)
            print(f"Validating ARM graph {ad}")
            plain_neo4j.validate_graph()

            site_arm = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=plain_neo4j.graph_id,
                                                              importer=self.n4j_imp))
            # generate a dict of ADMs from site graph ARM
            site_adms = site_arm.generate_adms()
            print('ADMS' + str(site_adms.keys()))

            # desired ADM is under 'primary'
            site_adm = site_adms['primary']
            cbm.merge_adm(adm=site_adm)

            print('Deleting ADM and ARM graphs')
            for adm in site_adms.values():
                adm_ids[ad] = adm.graph_id
                adm.delete_graph()
            site_arm.delete_graph()

        cbm.validate_graph()
        print('CBM ID is ' + cbm.graph_id)

        cbm_graph_id = cbm.graph_id
        # turn on debug so we can test formation of ABQM without querying
        # actor for reservations
        AggregatedBQMPlugin.DEBUG_FLAG = True

        n4j_pg = Neo4jPropertyGraph(graph_id=cbm_graph_id, importer=self.n4j_imp)

        cbm = Neo4jCBMFactory.create(n4j_pg)

        plugin = AggregatedBQMPlugin(actor=None, logger=None)

        abqm = plugin.plug_produce_bqm(cbm=cbm, query_level=1)

        abqm.validate_graph()

        abqm_string = abqm.serialize_graph()

        print('Writing ABQM to abqm.graphml')
        with open('abqm.graphml', 'w') as f:
            f.write(abqm_string)

        abqm_json = abqm.serialize_graph(format=GraphFormat.JSON_NODELINK)
        print('Writing ABQM to abqm.json')
        with open('abqm.json', 'w') as f:
            f.write(abqm_json)

        print('Writing CBM to cbm.graphml')
        cbm_string = cbm.serialize_graph()
        with open('cbm.graphml', 'w') as f:
            f.write(cbm_string)

        self.n4j_imp.delete_all_graphs()

    def test_cbm(self):
        self.n4j_imp.delete_all_graphs()
        # these are produced by substrate tests
        #cbm = '../../../neo4j/abqm-l2-cbm.graphml'
        cbm = 'cbm.graphml'

        plain_cbm = self.n4j_imp.import_graph_from_file_direct(graph_file=cbm)
        cbm = Neo4jCBMFactory.create(Neo4jPropertyGraph(graph_id=plain_cbm.graph_id,
                                     importer=self.n4j_imp))
        cbm.validate_graph()

        print('CBM ID is ' + cbm.graph_id)

        cbm_graph_id = cbm.graph_id
        # turn on debug so we can test formation of ABQM without querying
        # actor for reservations
        AggregatedBQMPlugin.DEBUG_FLAG = True

        n4j_pg = Neo4jPropertyGraph(graph_id=cbm_graph_id, importer=self.n4j_imp)

        cbm = Neo4jCBMFactory.create(n4j_pg)

        plugin = AggregatedBQMPlugin(actor=None, logger=None)

        abqm = plugin.plug_produce_bqm(cbm=cbm, query_level=1)

        abqm.validate_graph()

        abqm_string = abqm.serialize_graph()

        print('Writing ABQM to abqm-from-cbm.graphml')
        with open('abqm-from-cbm.graphml', 'w') as f:
            f.write(abqm_string)

        self.n4j_imp.delete_all_graphs()


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_mb.message_bus.producer import AvroProducerApi

from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation
from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
from fabric_cf.actor.core.apis.abc_client_callback_proxy import ABCClientCallbackProxy
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_rpc_request_state import ABCRPCRequestState
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.rpc_request_state import RPCRequestState
from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric_cf.actor.core.proxies.local.local_return import LocalReturn
from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.security.auth_token import AuthToken


class ClientCallbackHelper(ABCClientCallbackProxy):
    class MyRequestState(RPCRequestState):
        def __init__(self):
            super().__init__()
            self.reservation = None
            self.update_data = None

    def __init__(self, *, name: str, guid: ID):
        self.token = AuthToken(name=name, guid=guid)
        self.called = 0
        self.prepared = 0
        self.reservation = None

    def get_called(self) -> int:
        return self.called

    def get_guid(self) -> ID:
        return self.token.get_guid()

    def get_identity(self) -> AuthToken:
        return self.token

    def get_name(self) -> str:
        return self.token.get_name()

    def get_reservation(self) -> ABCReservationMixin:
        return self.reservation

    def get_type(self):
        return Constants.PROTOCOL_LOCAL

    def prepare_update_delegation(self, *, delegation: ABCDelegation, update_data: UpdateData,
                                  callback: ABCCallbackProxy, caller: AuthToken) -> ABCRPCRequestState:
        raise NotImplementedError

    def set_logger(self, *, logger):
        pass

    def prepare_update_ticket(self, *, reservation: ABCBrokerReservation, update_data: UpdateData,
                              callback: ABCCallbackProxy, caller: AuthToken) -> ABCRPCRequestState:
        state = self.MyRequestState()
        state.reservation = LocalReturn.pass_reservation(reservation=reservation,
                                                         plugin=ActorRegistrySingleton.get().get_actor(
                                                             actor_name_or_guid=self.token.get_name()).get_plugin())

        self.prepared += 1
        return state

    def execute(self, *, request: ABCRPCRequestState, producer: AvroProducerApi):
        if request.get_type() == RPCRequestType.UpdateTicket:
            self.called += 1
            self.reservation = request.reservation

    def get_logger(self):
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        return GlobalsSingleton.get().get_logger()

    def prepare_query_result(self, *, request_id: str, response, caller: AuthToken) -> ABCRPCRequestState:
        raise NotImplementedError

    def prepare_failed_request(self, *, request_id: str, failed_request_type,
                               failed_reservation_id: ID, error: str, caller: AuthToken) -> ABCRPCRequestState:
        raise NotImplementedError


# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server import util


class Slice(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, model: str=None, lease_start_time: str=None, lease_end_time: str=None, state: str=None, project_id: str=None, project_name: str=None, graph_id: str=None, name: str=None, slice_id: str=None):  # noqa: E501
        """Slice - a model defined in Swagger

        :param model: The model of this Slice.  # noqa: E501
        :type model: str
        :param lease_start_time: The lease_start_time of this Slice.  # noqa: E501
        :type lease_start_time: str
        :param lease_end_time: The lease_end_time of this Slice.  # noqa: E501
        :type lease_end_time: str
        :param state: The state of this Slice.  # noqa: E501
        :type state: str
        :param project_id: The project_id of this Slice.  # noqa: E501
        :type project_id: str
        :param project_name: The project_name of this Slice.  # noqa: E501
        :type project_name: str
        :param graph_id: The graph_id of this Slice.  # noqa: E501
        :type graph_id: str
        :param name: The name of this Slice.  # noqa: E501
        :type name: str
        :param slice_id: The slice_id of this Slice.  # noqa: E501
        :type slice_id: str
        """
        self.swagger_types = {
            'model': str,
            'lease_start_time': str,
            'lease_end_time': str,
            'state': str,
            'project_id': str,
            'project_name': str,
            'graph_id': str,
            'name': str,
            'slice_id': str
        }

        self.attribute_map = {
            'model': 'model',
            'lease_start_time': 'lease_start_time',
            'lease_end_time': 'lease_end_time',
            'state': 'state',
            'project_id': 'project_id',
            'project_name': 'project_name',
            'graph_id': 'graph_id',
            'name': 'name',
            'slice_id': 'slice_id'
        }
        self._model = model
        self._lease_start_time = lease_start_time
        self._lease_end_time = lease_end_time
        self._state = state
        self._project_id = project_id
        self._project_name = project_name
        self._graph_id = graph_id
        self._name = name
        self._slice_id = slice_id

    @classmethod
    def from_dict(cls, dikt) -> 'Slice':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The slice of this Slice.  # noqa: E501
        :rtype: Slice
        """
        return util.deserialize_model(dikt, cls)

    @property
    def model(self) -> str:
        """Gets the model of this Slice.


        :return: The model of this Slice.
        :rtype: str
        """
        return self._model

    @model.setter
    def model(self, model: str):
        """Sets the model of this Slice.


        :param model: The model of this Slice.
        :type model: str
        """

        self._model = model

    @property
    def lease_start_time(self) -> str:
        """Gets the lease_start_time of this Slice.


        :return: The lease_start_time of this Slice.
        :rtype: str
        """
        return self._lease_start_time

    @lease_start_time.setter
    def lease_start_time(self, lease_start_time: str):
        """Sets the lease_start_time of this Slice.


        :param lease_start_time: The lease_start_time of this Slice.
        :type lease_start_time: str
        """

        self._lease_start_time = lease_start_time

    @property
    def lease_end_time(self) -> str:
        """Gets the lease_end_time of this Slice.


        :return: The lease_end_time of this Slice.
        :rtype: str
        """
        return self._lease_end_time

    @lease_end_time.setter
    def lease_end_time(self, lease_end_time: str):
        """Sets the lease_end_time of this Slice.


        :param lease_end_time: The lease_end_time of this Slice.
        :type lease_end_time: str
        """

        self._lease_end_time = lease_end_time

    @property
    def state(self) -> str:
        """Gets the state of this Slice.


        :return: The state of this Slice.
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state: str):
        """Sets the state of this Slice.


        :param state: The state of this Slice.
        :type state: str
        """

        self._state = state

    @property
    def project_id(self) -> str:
        """Gets the project_id of this Slice.


        :return: The project_id of this Slice.
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id: str):
        """Sets the project_id of this Slice.


        :param project_id: The project_id of this Slice.
        :type project_id: str
        """

        self._project_id = project_id

    @property
    def project_name(self) -> str:
        """Gets the project_name of this Slice.


        :return: The project_name of this Slice.
        :rtype: str
        """
        return self._project_name

    @project_name.setter
    def project_name(self, project_name: str):
        """Sets the project_name of this Slice.


        :param project_name: The project_name of this Slice.
        :type project_name: str
        """

        self._project_name = project_name

    @property
    def graph_id(self) -> str:
        """Gets the graph_id of this Slice.


        :return: The graph_id of this Slice.
        :rtype: str
        """
        return self._graph_id

    @graph_id.setter
    def graph_id(self, graph_id: str):
        """Sets the graph_id of this Slice.


        :param graph_id: The graph_id of this Slice.
        :type graph_id: str
        """
        if graph_id is None:
            raise ValueError("Invalid value for `graph_id`, must not be `None`")  # noqa: E501

        self._graph_id = graph_id

    @property
    def name(self) -> str:
        """Gets the name of this Slice.


        :return: The name of this Slice.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Sets the name of this Slice.


        :param name: The name of this Slice.
        :type name: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def slice_id(self) -> str:
        """Gets the slice_id of this Slice.


        :return: The slice_id of this Slice.
        :rtype: str
        """
        return self._slice_id

    @slice_id.setter
    def slice_id(self, slice_id: str):
        """Sets the slice_id of this Slice.


        :param slice_id: The slice_id of this Slice.
        :type slice_id: str
        """
        if slice_id is None:
            raise ValueError("Invalid value for `slice_id`, must not be `None`")  # noqa: E501

        self._slice_id = slice_id
